from __future__ import print_function

from packaging import version

from ifigure.utils.cross2d import *
from ifigure.matplotlib_mod.canvas_common import camera_distance

import time

import matplotlib
from matplotlib.image import FigureImage
from functools import wraps
from ifigure.matplotlib_mod.is_supported_renderer import isSupportedRenderer
from matplotlib.collections import Collection, LineCollection, \
    PolyCollection, PatchCollection, PathCollection
import numpy as np
from matplotlib.artist import allow_rasterization
import weakref
import ifigure.events as events
from scipy.signal import convolve2d, fftconvolve
from scipy.sparse import coo_matrix

from matplotlib.axes import Axes
from matplotlib.lines import Line2D
from matplotlib.text import Text
from mpl_toolkits.mplot3d.axes3d import Axes3D
import mpl_toolkits.mplot3d.art3d as art3d
import matplotlib.transforms as trans
from matplotlib.colors import ColorConverter
from functools import reduce
cc = ColorConverter()


# KERNEL for mask bluring
conv_kernel_size = 11
x = 1 - np.abs(np.linspace(-1., 1., conv_kernel_size))
X, Y = np.meshgrid(x, x)
conv_kernel = np.sqrt(X**2, Y**2)
conv_kernel = conv_kernel / np.sum(conv_kernel)
###


def convert_to_gl(obj, zs=0, zdir='z'):
    from .art3d_gl import polygon_2d_to_gl
    from .art3d_gl import line_3d_to_gl
    from .art3d_gl import poly_collection_3d_to_gl
    from .art3d_gl import line_collection_3d_to_gl
    from matplotlib.patches import Polygon
    from mpl_toolkits.mplot3d.art3d import Poly3DCollection, Line3D, \
        poly_collection_2d_to_3d, Line3DCollection

    if isinstance(obj, Line3D):
        line3d_to_gl(obj)
    elif isinstance(obj, Polygon):
        polygon_2d_to_gl(obj, zs, zdir)
    elif isinstance(obj, Poly3DCollection):
        poly_collection_3d_to_gl(obj)
    elif isinstance(obj, Line3DCollection):
        line_collection_3d_to_gl(obj)
    elif isinstance(obj, PolyCollection):
        poly_collection_2d_to_3d(obj, zs=zs, zdir=zdir)
        poly_collection_3d_to_gl(obj)


def get_glcanvas():
    from ifigure.matplotlib_mod.backend_wxagg_gl import FigureCanvasWxAggModGL
    return FigureCanvasWxAggModGL.glcanvas


def norm_vec(n):
    d = np.sum(n**2)
    if d == 0:
        return np.zeros(len(n))
    else:
        return n / np.sqrt(d)


def arrow3d(base, r1, r2, ort, l, h, m=13, pivot='tail', d1d2=None):
    ort = norm_vec(ort)
    th = np.linspace(0, np.pi * 2, m).reshape(-1, 1)

    if d1d2 is None:
        x = np.array([1., 0., 0.])
        y = np.array([0., 1., 0.])
        z = np.array([0., 0., 1.])

        if np.sum(np.cross(ort, x)**2) == 0:
            if np.sum(np.cross(ort, y)**2) == 0:
                d1 = norm_vec(np.cross(ort, z))
            else:
                d1 = norm_vec(np.cross(ort, y))
        else:
            d1 = norm_vec(np.cross(ort, x))
        d2 = np.cross(ort, d1)
    else:
        d1, d2 = d1d2

    if pivot == 'tip':
        base = base - (l + h) * ort
    elif pivot == 'mid':
        base = base - (l + h) * ort / 2.
    else:
        pass

    p = base + l * r1 * (d1 * np.cos(th) + d2 * np.sin(th))
    q = p + l * ort
    p2 = base + l * r2 * (d1 * np.cos(th) + d2 * np.sin(th)) + l * ort
    p3 = base + (l + h) * ort
    p3 = np.array([p3] * m).reshape(-1, 3)
    t1 = np.stack((p[:-1], q[:-1], p[1:]), axis=1)
    t2 = np.stack((p[1:], q[:-1], q[1:]), axis=1)
    t3 = np.stack((p2[:-1], p3[:-1], p2[1:]), axis=1)
    # t2 = np.dstack((p[1:], q[:-1], q[1:]))
    t1 = np.vstack((t1, t2, t3))
    return t1


def world_transformation(xmin, xmax,
                         ymin, ymax,
                         zmin, zmax,
                         view_scale=1):
    dx, dy, dz = (xmax - xmin), (ymax - ymin), (zmax - zmin)
    return np.array([
        [1.0 / dx, 0, 0, -xmin / dx],
        [0, 1.0 / dy, 0, -ymin / dy],
        [0, 0, 1.0 / dz, -zmin / dz],
        [0, 0, 0, 1.0]])


def view_transformation(E, R, V):
    n = (E - R)
    # new
#    n /= mod(n)
#    u = np.cross(V,n)
#    u /= mod(u)
#    v = np.cross(n,u)
#    Mr = np.diag([1.]*4)
#    Mt = np.diag([1.]*4)
#    Mr[:3,:3] = u,v,n
#    Mt[:3,-1] = -E
    # end new

    # old
    n = n / np.sqrt(np.sum(n**2))
    u = np.cross(V, n)
    u = u / np.sqrt(np.sum(u**2))
    v = np.cross(n, u)
    Mr = [[u[0], u[1], u[2], 0],
          [v[0], v[1], v[2], 0],
          [n[0], n[1], n[2], 0],
          [0, 0, 0, 1],
          ]
    #
    Mt = [[1, 0, 0, -E[0]],
          [0, 1, 0, -E[1]],
          [0, 0, 1, -E[2]],
          [0, 0, 0, 1]]
    # end old

    return np.dot(Mr, Mt)


def persp_transformation(zfront, zback):
    a = (zfront + zback) / (zfront - zback)
    b = -2 * (zfront * zback) / (zfront - zback)
    from ifigure.matplotlib_mod.canvas_common import view_scale
    return np.array([[view_scale, 0, 0, 0],
                     [0, view_scale, 0, 0],
                     [0, 0, a, b],
                     [0, 0, -1, 0]
                     ])


def rotation_mat(ax, an):
    '''
    matrix to rotate arounc ax by an [rad]
    '''
    c = np.cos(an)
    s = np.sin(an)
    ax = ax / np.sqrt(np.sum(ax**2))
    R = np.array([[c + (1 - c) * ax[0]**2,
                   ax[0] * ax[1] * (1 - c) - ax[2] * s,
                   ax[0] * ax[2] * (1 - c) + ax[1] * s],
                  [ax[0] * ax[1] * (1 - c) + ax[2] * s,
                   c + (1 - c) * ax[1]**2,
                   ax[1] * ax[2] * (1 - c) - ax[0] * s],
                  [ax[0] * ax[2] * (1 - c) - ax[1] * s,
                   ax[1] * ax[2] * (1 - c) + ax[0] * s,
                   c + (1 - c) * ax[2]**2]])

    return R


def use_gl_switch(func):
    '''
    use_gl_switch allows to select types of 3D plot
    aritist, either mplot3d or openGL based artist.

    note that piScope does not keep track
    use_gl switch. Manipulating use_gl manually in
    piScope makes your plot unreproducible.
    '''
    @wraps(func)
    def checker(self, *args, **kargs):
        if self._use_gl:
            m = func
            ret = m(self, *args, **kargs)
        else:
            m = getattr(super(Axes3DMod, self), func.__name__)
            ret = m(*args, **kargs)
        return ret
    return checker


class ArtGLHighlight(FigureImage):
    def remove(self):
        self.figure.artists.remove(self)


class Axes3DMod(Axes3D):
    pan_sensitivity = 5

    def __init__(self, *args, **kargs):
        self._nomargin_mode = False
        self._offset_trans_changed = False
        self._mouse_hit = False
        self._lighting = {'light': 0.5,
                          'ambient': 0.5,
                          'specular': .0,
                          'light_direction': (1., 0, 1, 0),
                          'light_color': (1., 1., 1),
                          'wireframe': 0,
                          'clip_limit1': [1., 0., 0.],
                          'clip_limit2': [0., 1., 1.],
                          'shadowmap': False}
        self._use_gl = kargs.pop('use_gl', True)
        self._use_frustum = kargs.pop('use_frustum', True)
        self._use_clip = kargs.pop('use_clip', 1)

        if version.parse(matplotlib.__version__) >= version.parse("3.4"):
            kargs['auto_add_to_figure'] = False
            super(Axes3DMod, self).__init__(*args, **kargs)
            args[0].add_axes(self)
        else:
            super(Axes3DMod, self).__init__(*args, **kargs)

        self.patch.set_alpha(0)
        self._gl_id_data = None
        self._gl_mask_artist = None
        self._3d_axes_icon = None
        self._show_3d_axes = True
        self._upvec = np.array([0, 0, 1])
        self._ignore_screen_aspect_ratio = True
        self._gl_scale = 1.0
        self._gl_scale_accum = 1.0
        self._gl_rot_center = np.array([0., 0., 0.])
        self._gl_use_rot_center = False
        self._zoom_btn = []
        self._pan_btn = []
        self._rotate_btn = []
        self._drag_mode = ''

        self._sx, self._sy = 0, 0

        from ifigure.matplotlib_mod.axis3d_mod import XAxis, YAxis, ZAxis
        self.xaxis.__class__ = XAxis
        self.yaxis.__class__ = YAxis
        self.zaxis.__class__ = ZAxis

    @property
    def dist(self):
        return camera_distance

    @dist.setter
    def dist(self, value):
        pass

    # after MPL 3.6.1 dist became _dist
    @property
    def _dist(self):
        return camera_distance

    @_dist.setter
    def _dist(self, value):
        pass

    # after MPL 3.6.1, initial_roll is defined
    @property
    def _initial_roll(self):
        if hasattr(self, "initial_roll"):
            return self.initial_roll
        else:
            return (0, 0, 1)

    def view_init(self, elev=None, azim=None, roll=None, vertical_axis="z"):
        """
        Copied form Axes3D to play with self.dist
        """
        # self._dist = camera_distance

        if elev is None:
            self.elev = self.initial_elev
        else:
            self.elev = elev

        if azim is None:
            self.azim = self.initial_azim
        else:
            self.azim = azim

        if roll is None:
            self.roll = self._initial_roll
        else:
            self.roll = roll

        from matplotlib import _api
        self._vertical_axis = _api.check_getitem(
            dict(x=0, y=1, z=2), vertical_axis=vertical_axis
        )

    def gl_hit_test(self, x, y, artist, radius=3):
        #
        #  logic is
        #     if artist_id is found within raidus from (x, y)
        #     and
        #     if it is the closet artist in the area of checking
        #     then return True
        if self._gl_id_data is None:
            return False, None

        x0, y0, id_dict, im, imd, im2 = self._gl_id_data

        x, x0, y, y0 = int(x), int(x0), int(y), int(y0)

        d = (im[y - y0 - radius:y - y0 + radius,
                x - x0 - radius:x - x0 + radius]).flatten()

        dd = (im2[y - y0 - radius:y - y0 + radius,
                  x - x0 - radius:x - x0 + radius]).flatten()
        dd_extra = (imd[y - y0 - radius:y - y0 + radius,
                        x - x0 - radius:x - x0 + radius]).flatten()

        if len(dd) == 0:
            return False, None

        mask = np.array([id_dict[x]()._gl_pickable if x in id_dict
                         and id_dict[x]() is not None else False
                         for x in d])
        if not any(mask):
            return False, None

        dist = np.min(dd[mask])

        for num, check, check2 in zip(d, dd, dd_extra):
            if num in id_dict:
                if id_dict[num]() == artist and check == dist:
                    return True, check2

        return False, None

    def gl_hit_test_rect(self, rect, artist, check_selected_all_covered=False):
        #
        #     if artist_id is found within raidus from (x, y)
        #     and
        #     if it is the closet artist in the area of checking
        #     then return True
        if self._gl_id_data is None:
            return False, False, None
        if not artist._gl_pickable:
            return False, False, None

        x0, y0, id_dict, im, imd, im2 = self._gl_id_data

        x, x0, y, y0 = int(rect[0]), int(x0), int(rect[1]), int(y0)
        dx = rect[2]
        dy = rect[3]

        d = (im[y - y0:y - y0 + dy, x - x0:x - x0 + dx]).flatten()
        dd_extra = (imd[y - y0:y - y0 + dy, x - x0:x - x0 + dx]).flatten()

        if len(d) == 0:
            return False, False, None

        xxx = -1
        for key in id_dict:
            if id_dict[key]() is None:
                continue
            if id_dict[key]() == artist:
                xxx = key
                break

        mask = (d == key)
        mask_all = (im == key)
        all_covered = (np.sum(mask) == np.sum(mask_all))

        if not any(mask):
            return False, False, None

        dd_extra1 = dd_extra[mask]
        dd_extra2 = imd[mask_all]

        selected_idx = np.unique(dd_extra1)

        if check_selected_all_covered:
            all_covered_selected_idx = []
            for i in selected_idx:
                if np.sum(dd_extra1 == i) == np.sum(dd_extra2 == i):
                    all_covered_selected_idx.append(i)
            selected_idx = all_covered_selected_idx

        return True, all_covered, selected_idx

    def make_gl_hl_artist(self):
        if self._gl_id_data is None:
            return []
        self.del_gl_hl_artist()

        x0, y0, id_dict, im, imd, im2 = self._gl_id_data

        a = ArtGLHighlight(self.figure, offsetx=x0,
                           offsety=y0, origin='lower')
        data = np.ones(im.shape)
        alpha = np.zeros(im.shape)
        mask = np.dstack((data, data, data, alpha))
        a.set_array(mask)
        self.figure.artists.append(a)
        self._gl_mask_artist = a

        return [a]

    def del_gl_hl_artist(self):
        if self._gl_mask_artist is not None:
            self._gl_mask_artist.remove()
        self._gl_mask_artist = None

    def set_gl_hl_mask(self, artist, hit_id=None,
                       cmask=0.0, amask=0.65):
        #
        #  logic is
        #     if artist_id is found within raidus from (x, y)
        #     and
        #     if it is the closet artist in the area of checking
        #     then return True

        if self._gl_id_data is None:
            return False

        if self._gl_mask_artist is None:
            return False

        # do not do this when hitest_map is updating..this is when
        # mouse dragging is going on
        if not get_glcanvas()._hittest_map_update:
            return

        x0, y0, id_dict, im, imd, im2 = self._gl_id_data
        arr = self._gl_mask_artist.get_array()

        for k in id_dict:
            if (id_dict[k]() == artist):
                if hit_id is not None:
                    if len(hit_id) > 0:
                        mask = np.isin(imd, hit_id)
                        m = np.logical_and(im == k, mask)
                    else:
                        m = (im == k)
                else:
                    m = (im == k)

                c = self.figure.canvas.hl_color
                arr[:, :, :3][m] = np.array(c, copy=None)
                arr[:, :, 3][m] = amask
                break
        # blur the mask,,,

    def blur_gl_hl_mask(self, amask=0.65):
        if self._gl_mask_artist is None:
            return
        arr = self._gl_mask_artist.get_array()
        # b = convolve2d(arr[:,:,3], conv_kernel, mode = 'same') + arr[:,:,3]
        # b = fftconvolve(arr[:, :, 3], conv_kernel, mode='same') + arr[:, :, 3]
        from scipy.ndimage import gaussian_filter
        b = gaussian_filter(arr[:, :, 3], sigma=1) + arr[:, :, 3]
        # b = np.sqrt(b)
        b[b > amask] = amask

        c = self.figure.canvas.hl_color
        arr[b > 0.0, :3] = c
        arr[..., -1] = b
        self._gl_mask_artist.set_array(arr)

    def set_nomargin_mode(self, mode):
        self._nomargin_mode = mode

    def get_nomargin_mode(self):
        return self._nomargin_mode

    def get_zaxis(self):
        return self.zaxis

    #
    #   prepend and append some work to visual
    #   effect for 3D rotation
    #
    def set_mouse_button(self, rotate_btn=1, zoom_btn=3, pan_btn=2):
        self._pan_btn = np.atleast_1d(pan_btn)
        self._rotate_btn = np.atleast_1d(rotate_btn)
        # self._zoom_btn = np.atleast_1d(zoom_btn)
        self._zoom_btn = []

    def _button_press(self, evt):
        self._mouse_hit, extra = self.contains(evt)
        if not self._mouse_hit:
            return
        fig_axes = self.figobj
#        for obj in fig_axes.walk_tree():
#            obj.switch_scale('coarse')
        Axes3D._button_press(self, evt)

    def _on_move_start(self):
        get_glcanvas()._hittest_map_update = False

    def _on_move(self, evt):
        if not self._mouse_hit:
            return

        fig_axes = self.figobj
        if fig_axes is None:
            return

        fig_axes.set_bmp_update(False)

        self._on_move_mod(evt)
        get_glcanvas()._hittest_map_update = False

        if self._drag_mode in ['zoom', 'select']:
            return

        events.SendPVDrawRequest(self.figobj,
                                 w=None, wait_idle=False,
                                 refresh_hl=False,
                                 caller='_on_move')

    def _on_move_done(self):
        get_glcanvas()._hittest_map_update = True

    def calc_range_change_zoom3d(self, xdata, ydata, sxdata, sydata, updown):
        # this is to debug this method...
        # from ifigure.matplotlib_mod.calc_range_change_by_pan_test import test
        # return test(self, xdata, ydata, sxdata, sydata)
        dx = - (xdata + sxdata) / 2.0
        dy = - (ydata + sydata) / 2.0
        dp = np.array((dx, dy))

        w = self._pseudo_w
        h = self._pseudo_h

        # dx, dy : normalized screen coordinate of center of zoom
        df = max(abs(xdata - sxdata) / w, abs(ydata - sydata) / h, 0.05)
        if updown == 'down':
            df = 1. / df
        # 3D axes range
        minx, maxx, miny, maxy, minz, maxz = self.get_w_lims()

        center_move = self._screen_move_to_3d_move(-dp)
        minp = np.array([minx, miny, minz])
        maxp = np.array([maxx, maxy, maxz])

        r1 = (minp + maxp) / 2.0 - (maxp - minp) * df / 2 + center_move
        r2 = (minp + maxp) / 2.0 + (maxp - minp) * df / 2 + center_move

        return (r1[0], r2[0]), (r1[1], r2[1]), (r1[2], r2[2]), df

    def calc_range_change_wheel(self, xdata, ydata, sxdata, sydata, updown):
        # this is to debug this method...
        # from ifigure.matplotlib_mod.calc_range_change_by_pan_test import test
        # return test(self, xdata, ydata, sxdata, sydata)

        r = -0.15 if updown == 'down' else 0.15
        dx = - (xdata + sxdata) / 2.0 * r
        dy = - (ydata + sydata) / 2.0 * r
        dp = np.array((dx, dy))

        w = self._pseudo_w
        h = self._pseudo_h

        # dx, dy : normalized screen coordinate of center of zoom
        df = 1 / 0.95 if updown == 'down' else 0.95
        # 3D axes range
        minx, maxx, miny, maxy, minz, maxz = self.get_w_lims()

        center_move = self._screen_move_to_3d_move(-dp)
        minp = np.array([minx, miny, minz])
        maxp = np.array([maxx, maxy, maxz])

        r1 = (minp + maxp) / 2.0 - (maxp - minp) * df / 2 + center_move
        r2 = (minp + maxp) / 2.0 + (maxp - minp) * df / 2 + center_move

        return (r1[0], r2[0]), (r1[1], r2[1]), (r1[2], r2[2]), df

    def _on_move_mod(self, event):
        """
        added pan mode
        """
        if not self.button_pressed:
            return

        if self.M is None:
            return

        x, y = event.xdata, event.ydata
        # In case the mouse is out of bounds.
        if x is None:
            return

        dx, dy = x - self._sx, y - self._sy
        w = self._pseudo_w
        h = self._pseudo_h
        self._sx, self._sy = x, y

        # Rotation
        if self.button_pressed in self._rotate_btn:
            # rotate viewing point
            # get the x and y pixel coords
            if dx == 0 and dy == 0:
                return
            relev, razim = np.pi * self.elev / 180, np.pi * self.azim / 180
            p1 = np.array((np.cos(razim) * np.cos(relev),
                           np.sin(razim) * np.cos(relev),
                           np.sin(relev)))
            rightvec = np.cross(self._upvec, p1)

            newp1 = p1 - (dx / w * rightvec + dy / h * self._upvec) * \
                Axes3DMod.pan_sensitivity * self._gl_scale_accum

            newp1 = newp1 / np.sqrt(np.sum(newp1**2))
            self._upvec = self._upvec - newp1 * np.sum(newp1 * self._upvec)
            self._upvec = self._upvec / np.sqrt(np.sum(self._upvec**2))
            self.elev = np.arctan2(newp1[2], np.sqrt(
                newp1[0]**2 + newp1[1]**2)) * 180 / np.pi
            self.azim = np.arctan2(newp1[1], newp1[0]) * 180 / np.pi

            if self._gl_use_rot_center:
                Mold = self._matrix_cache_combined
                self.get_proj2()
                Mnew = self._matrix_cache_combined

                xrange, yrange, zrange = self._screen_shift_3drot(Mold, Mnew)
                self.set_xlim3d(xrange)
                self.set_ylim3d(yrange)
                self.set_zlim3d(zrange)

        elif self.button_pressed in self._pan_btn:
            dx = 1 - ((w - dx) / w)
            dy = 1 - ((h - dy) / h)
            relev, razim = np.pi * self.elev / 180, np.pi * self.azim / 180
            p1 = np.array((np.cos(razim) * np.cos(relev),
                           np.sin(razim) * np.cos(relev),
                           np.sin(relev)))
            rightvec = np.cross(self._upvec, p1)  # right on screen

            # p2 = np.array((np.sin(razim), -np.cos(razim), 0))
            # p3 = np.cross(p1, p2)
            # dx, dy, dz = p2*dx + p3*dy
            dx, dy, dz = -rightvec * dx - self._upvec * dy
            minx, maxx, miny, maxy, minz, maxz = self.get_w_lims()
            dx = (maxx - minx) * dx
            dy = (maxy - miny) * dy
            dz = (maxz - minz) * dz

            self.set_xlim3d(minx + dx, maxx + dx)
            self.set_ylim3d(miny + dy, maxy + dy)
            self.set_zlim3d(minz + dz, maxz + dz)

        # Zoom
        elif self.button_pressed in self._zoom_btn:
            # print("in zoom")
            # zoom view
            # hmmm..this needs some help from clipping....
            # this section not used..?
            minx, maxx, miny, maxy, minz, maxz = self.get_w_lims()
            df = 1 - ((h - dy) / h)
            dx = (maxx - minx) * df
            dy = (maxy - miny) * df
            dz = (maxz - minz) * df
            self.set_xlim3d(minx - dx, maxx + dx)
            self.set_ylim3d(miny - dy, maxy + dy)
            self.set_zlim3d(minz - dz, maxz + dz)
            # self.get_proj()
            # self.figure.canvas.draw_idle()

    def _button_release(self, evt):
        if not self._mouse_hit:
            return

        fig_axes = self.figobj
        if fig_axes is None:
            return

        fig_axes.set_bmp_update(False)
        Axes3D._button_release(self, evt)
        self._pan_btn = []
        self._rotate_btn = []
        self._zoom_btn = []
        self._mouse_hit = False
        get_glcanvas()._hittest_map_update = True

        # events.SendPVDrawRequest(self.figobj,
        #                         w=None, wait_idle=False,
        #                         refresh_hl=False,
        #                         caller = '_on_release')

    @use_gl_switch
    def plot(self, *args, **kwargs):
        from .art3d_gl import line_3d_to_gl
        fc = kwargs.pop('facecolor', None)
        gl_offset = kwargs.pop('gl_offset', (0, 0, 0))
        array_idx = kwargs.pop('array_idx', None)
        lines = Axes3D.plot(self, *args, **kwargs)
        for l in lines:
            line_3d_to_gl(l)
            l._facecolor = fc
            l._gl_offset = gl_offset
            l._gl_array_idx = array_idx
        return lines

    def fill(self, *args, **kwargs):
        from .art3d_gl import polygon_2d_to_gl
        zs = kwargs.pop('zs', 0)
        zdir = kwargs.pop('zdir', 'z')
        a = Axes3D.fill(self, *args, **kwargs)

        for obj in a:
            convert_to_gl(obj, zs, zdir)
        return a

    def fill_between(self, *args, **kwargs):
        from .art3d_gl import polygon_2d_to_gl
        zs = kwargs.pop('zs', 0)
        zdir = kwargs.pop('zdir', 'z')
        a = Axes3D.fill_between(self, *args, **kwargs)
        convert_to_gl(a, zs, zdir)
        a.convert_2dpath_to_3dpath(zs, zdir=zdir)
        return a

    def fill_betweenx(self, *args, **kwargs):
        from .art3d_gl import polygon_2d_to_gl
        zs = kwargs.pop('zs', 0)
        zdir = kwargs.pop('zdir', 'z')
        a = Axes3D.fill_betweenx(self, *args, **kwargs)
        convert_to_gl(a, zs, zdir)
        a.convert_2dpath_to_3dpath(zs, zdir=zdir)
        return a

    def cz_plot(self, x, y, z, c, **kywds):
        from ifigure.matplotlib_mod.art3d_gl import Line3DCollectionGL
        a = Line3DCollectionGL([], c_data=c, gl_lighting=False, **kywds)
        a._segments3d = (np.transpose(
            np.vstack((np.array(x), np.array(y), np.array(z)))),)
        a.convert_2dpath_to_3dpath()
        a.set_alpha(1.0)
        self.add_collection(a)
        return a

    def contour(self, *args, **kwargs):
        from .art3d_gl import poly_collection_3d_to_gl
        offset = kwargs['offset'] if 'offset' in kwargs else None
        zdir = kwargs['zdir'] if 'zdir' in kwargs else 'z'
        cset = Axes3D.contour(self, *args, **kwargs)
        '''
        for z, linec in zip(np.argsort(cset.levels), cset.collections):
            convert_to_gl(linec)
            linec.convert_2dpath_to_3dpath(z, zdir='z')
            linec.do_stencil_test = True
            if offset is not None:
                if zdir == 'x':
                    linec._gl_offset = (z * 0.001, 0, 0)
                elif zdir == 'y':
                    linec._gl_offset = (0, z * 0.001, 0)
                else:
                    linec._gl_offset = (0, 0, z * 0.001)
        '''
        return cset

    def imshow(self, *args, **kwargs):
        im_center = kwargs.pop('im_center', (0, 0, 0))
        im_axes = kwargs.pop('im_axes', [(1, 0, 0), (0, 1, 0)])

        from .art3d_gl import image_to_gl
        im = Axes3D.imshow(self, *args, **kwargs)
        image_to_gl(im)
        if len(args) > 2:
            extent = (min(args[0]), max(args[0]), min(args[1]), max(args[0]))
            im.set_extent(extent)
        im.set_im3dpath(im_center, im_axes)

        return im

    def contourf(self, *args, **kwargs):
        from .art3d_gl import poly_collection_3d_to_gl
        offset = kwargs['offset'] if 'offset' in kwargs else None
        zdir = kwargs['zdir'] if 'zdir' in kwargs else 'z'
        cset = Axes3D.contourf(self, *args, **kwargs)
        edgecolor = kwargs.pop('edgecolor', [1, 1, 1, 0])
        '''
        for z, linec in zip(np.argsort(cset.levels), cset.collections):
            convert_to_gl(linec)
            linec.convert_2dpath_to_3dpath(z, zdir='z')
            linec.do_stencil_test = True
            if offset is not None:
                if zdir == 'x':
                    linec._gl_offset = (z * 0.001, 0, 0)
                elif zdir == 'y':
                    linec._gl_offset = (0, z * 0.001, 0)
                else:
                    linec._gl_offset = (0, 0, z * 0.001)
            linec.set_edgecolor((edgecolor,))
        '''
        return cset

    def quiver(self, *args, **kwargs):
        '''
         quiver(x, y, z, u, v, w, length=0.1, normalize = True, **kwargs)

            kwargs: facecolor
                    edgecolor
                    alpha
                    cz, cdata

        '''
        # made based on mplot3d but makes GL solid object
        # handle kwargs
        # shaft length
        length = kwargs.pop('length', 1.0)
        # arrow length ratio to the shaft length
        arrow_length_ratio = kwargs.pop('arrow_length_ratio', 0.3)
        # pivot point (not implemeted)
        pivot = kwargs.pop('pivot', 'tail')
        # normalize
        normalize = kwargs.pop('normalize', False)

        # handle args
        argi = 6
        if len(args) < argi:
            ValueError('Wrong number of arguments. Expected %d got %d' %
                       (argi, len(args)))

        # first 6 arguments are X, Y, Z, U, V, W
        input_args = args[:argi]
        # if any of the args are scalar, convert into list
        input_args = [[k] if isinstance(k, (int, float)) else k
                      for k in input_args]

        # extract the masks, if any
        masks = [k.mask for k in input_args if isinstance(
            k, np.ma.MaskedArray)]
        # broadcast to match the shape
        bcast = np.broadcast_arrays(*(input_args + masks))
        input_args = bcast[:argi]
        masks = bcast[argi:]
        if masks:
            # combine the masks into one
            mask = reduce(np.logical_or, masks)
            # put mask on and compress
            input_args = [np.ma.array(k, mask=mask).compressed()
                          for k in input_args]
        else:
            input_args = [k.flatten() for k in input_args]
        XYZ = np.column_stack(input_args[:3])
        UVW = np.column_stack(input_args[3:argi]).astype(float)

        norm = np.sqrt(np.sum(UVW**2, axis=1))
        # If any row of UVW is all zeros, don't make a quiver for it
        mask = norm > 0
        norm = norm[mask]
        XYZ = XYZ[mask]
        ORT = UVW[mask] / norm.reshape((-1, 1))
        if normalize:
            norm = np.array([length] * len(ORT))
        else:
            norm = norm / np.max(norm) * length

        h = np.max(norm) * arrow_length_ratio
        r1 = kwargs.pop('shaftsize', 0.05)
        r2 = kwargs.pop('headsize', 0.25)

        m = 13
        sample_len = len(
            arrow3d(XYZ[0], r1, r2, ORT[0], norm[0], h, m=m, pivot=pivot))
        v = np.vstack([arrow3d(base, r1, r2, ort, l, h, m=m, pivot=pivot)
                       for base, ort, l in zip(XYZ, ORT, norm)],)
        cdata = kwargs.pop('facecolordata', None)
        if cdata is not None:
            cdata = np.transpose(
                np.vstack([cdata.flatten()] * sample_len)).flatten()
            kwargs['facecolordata'] = cdata

        return self.plot_solid(v, **kwargs)

    def plot_revolve(self, R, Z, *args, **kwargs):
        '''
        revolve

        '''
        raxis = np.array(kwargs.pop('raxis', (0, 1)))
        rtheta = kwargs.pop('rtheta', (0, np.pi * 2))
        rmesh = kwargs.pop('rmesh', 30)
        rcenter = np.array(kwargs.pop('rcenter', [0, 0]))
        theta = np.linspace(rtheta[0], rtheta[1], rmesh)

        pos = np.vstack((R - rcenter[0], Z - rcenter[1]))
        nraxis = raxis / np.sqrt(np.sum(raxis**2))
        nraxis = np.hstack([nraxis.reshape(2, -1)] * len(R))
        nc = np.hstack([rcenter.reshape(2, -1)] * len(R))
        dcos = np.sum(pos * nraxis, 0)
        newz = nc + dcos  # center of rotation
        dsin = pos[0, :] * nraxis[1, :] - pos[1, :] * nraxis[0, :]

#        Theta, R = np.meshgrid(theta, np.abs(dsin))
#        void, Z = np.meshgrid(theta, dcos)
        R, Theta = np.meshgrid(np.abs(dsin), theta)
        Z, void = np.meshgrid(dcos, theta)

        X = R * np.cos(Theta)
        Y = R * np.sin(Theta)

        tt = np.pi / 2 - np.arctan2(raxis[1], raxis[0])
        m = np.array([[np.cos(tt), 0, -np.sin(tt)],
                      [0, 1, 0],
                      [np.sin(tt), 0, np.cos(tt)], ])

        dd = np.dot(np.dstack((X, Y, Z)), m)
        X = dd[:, :, 0] + rcenter[0]
        Y = dd[:, :, 1]
        Z = dd[:, :, 2] + rcenter[1]

#        from ifigure.interactive import figure
#        v = figure()
#        v.plot(X, Z)
        # facecolor = kwargs.pop('facecolor', (0,0,1,1))
        X, Y, Z = np.broadcast_arrays(X, Y, Z)
        polyc = self.plot_surface(X, Y, Z, *args, **kwargs)
        polyc._revolve_data = (X, Y, Z)
        return polyc

    def plot_extrude(self, X, Y, Z, path,
                     scale=None, revolve=False):
        '''
        extrude a path drawn by X, Y, Z along the path
            path.shape = [:, 3]

         A, B = np.meshgrid(ai, bj)
           A.shape = (j, i). A[j,i] = ai
           B.shape = (j, i). B[j,i] = bj

           X[j, i] = X[i] - path_x[0] + path_x[j]
           X.flatten() = X[0, :], X[1,:], X[2,:]
        '''
        facecolor = kwargs.pop('facecolor', (0, 0, 1, 1))
        scale = kwargs.pop('scale', 1.0)
        scale = kwargs.pop('scale', 1.0)
        x1, x2 = np.meshgrid(path[:, 0], X)
        x = x1 + x2 - path[0, 0]
        y1, y2 = np.meshgrid(path[:, 1], Y)
        y = y1 + y2 - path[0, 1]
        z1, z2 = np.meshgrid(path[:, 2], Z)
        z = z1 + z2 - path[0, 2]

        X = x.flatten()
        Y = y.flatten()
        Z = z.flatten()
        polyc = self.plot_surface(X, Y, Z, *args, **kwargs)
        return polyc

    def plot_surface(self, X, Y, Z, *args, **kwargs):
        '''
        Create a surface plot using OpenGL-based artist

        By default it will be colored in shades of a solid color,
        but it also supports color mapping by supplying the *cmap*
        argument.

        ============= ================================================
        Argument      Description
        ============= ================================================
        *X*, *Y*, *Z* Data values as 2D arrays
        *edgecolor*   Color of the surface patches (default 'k')
        *facecolor*   Color of the surface patches (default None: use cmap)
        *rstride*     Reduce data
        *cstride*     Reduce data
        *cmap*        A colormap for the surface patches.
        *shade*       Whether to shade the facecolors
        ============= ================================================

        Other arguments are passed on to
        :class:`~mpl_toolkits.mplot3d.art3d.Poly3DCollection`
        '''
        cz = kwargs.pop('cz', False)
        cdata = kwargs.pop('cdata', None)
        expanddata = kwargs.pop('expanddata', False)

        Z = np.atleast_2d(Z)
        # TODO: Support masked arrays
        if Y.ndim == 1 and X.ndim == 1:
            X, Y = np.meshgrid(X, Y)
        X, Y, Z = np.broadcast_arrays(X, Y, Z)
        rows, cols = Z.shape

        rstride = kwargs.pop('rstride', 10)
        cstride = kwargs.pop('cstride', 10)
        idxset3d = []

        r = list(range(0, rows, rstride))
        c = list(range(0, cols, cstride))

        X3D = X[r, :][:, c].flatten()
        Y3D = Y[r, :][:, c].flatten()
        Z3D = Z[r, :][:, c].flatten()

        # array index
        idxset = []
        l_r = len(r)
        l_c = len(c)
#        offset = np.array([0, 1, l_c+1, l_c, 0])
        offset = np.array([0, 1, l_c + 1, l_c])
        base = np.arange(l_r * l_c).reshape(l_r, l_c)
        base = base[:-1, :-1].flatten()

        idxset = np.array([x + offset for x in base], 'H')

        if expanddata:
            verts = np.dstack((X3D[idxset],
                               Y3D[idxset],
                               Z3D[idxset]))
            if cz:
                if cdata is not None:
                    cdata = cdata[r, :][:, c].flatten()[idxset]
                else:
                    cdata = Z3D[idxset]
                shade = kwargs.pop('shade', 'flat')
                if shade != 'linear':
                    cdata = np.mean(cdata, -1)
                kwargs['facecolordata'] = np.real(cdata)
                kwargs.pop('facecolor', None)  # get rid of this keyword
            kwargs['cz'] = cz
            o = self.plot_solid(verts, **kwargs)
            o._idxset = (r, c, idxset)   # this is used for phasor
        else:
            verts = np.vstack((X3D, Y3D, Z3D)).transpose()
            if cz:
                if cdata is not None:
                    cdata = cdata[r, :][:, c].flatten()
                else:
                    cdata = Z3D
                shade = kwargs.get('shade', 'flat')
                kwargs['facecolordata'] = np.real(cdata)
                kwargs.pop('facecolor', None)  # get rid of this keyword
            kwargs['cz'] = cz
            o = self.plot_solid(verts, idxset, **kwargs)
            o._idxset = (r, c, None)   # this is used for phasor
        return o

    def plot_trisurf(self, *args, **kwargs):
        '''
        plot_trisurf(x, y, z,  **wrargs)
        plot_trisurf(x, y, z,  triangles = triangle,,,)
        plot_trisurf(tri, z,  **kwargs, cz = cz, cdata = cdata)


        '''
        from .art3d_gl import poly_collection_3d_to_gl
        from matplotlib.tri import Triangulation

        cz = kwargs.pop('cz', False)
        cdata = kwargs.pop('cdata', None)
        expanddata = kwargs.pop('expanddata', False)

        tri, args, kwargs = Triangulation.get_from_args_and_kwargs(
            *args, **kwargs)
        if 'Z' in kwargs:
            z = np.asarray(kwargs.pop('Z'))
        else:
            z = np.asarray(args[0])
            # We do this so Z doesn't get passed as an arg to PolyCollection
            args = args[1:]

        triangles = tri.get_masked_triangles()
        X3D = tri.x
        Y3D = tri.y
        Z3D = z
        idxset = tri.get_masked_triangles()

        if expanddata:
            verts = np.dstack((X3D[idxset],
                               Y3D[idxset],
                               Z3D[idxset]))
            if cz:
                if cdata is not None:
                    cdata = cdata[idxset]
                else:
                    cdata = Z3D[idxset]
                shade = kwargs.pop('shade', 'linear')
                if shade != 'linear':
                    cdata = np.mean(cdata, -1)
                kwargs['facecolordata'] = np.real(cdata)
                kwargs.pop('facecolor', None)  # get rid of this keyword
            kwargs['cz'] = cz
            o = self.plot_solid(verts, **kwargs)
            o._idxset = (None, None, idxset)   # this is used for phasor
        else:
            verts = np.vstack((X3D, Y3D, Z3D)).transpose()
            if cz:
                if cdata is not None:
                    cdata = cdata
                else:
                    cdata = Z3D
                kwargs['facecolordata'] = np.real(cdata)
                kwargs.pop('facecolor', None)  # get rid of this keyword
            kwargs['cz'] = cz
            o = self.plot_solid(verts, idxset, **kwargs)
            o._idxset = (None, None, None)   # this is used for phasor
        return o

    def prep_flat_shading_data(self, args, kwargs):
        if len(args) < 2:
            return args, kwargs

        vert = args[0]
        idx = args[1]
        args = (vert[idx],)
        if 'array_idx' in kwargs:
            if kwargs['array_idx'] is not None:
                kwargs['array_idx'] = kwargs['array_idx'][idx].flatten()
        if 'facecolordata' in kwargs:
            cdata = kwargs['facecolordata'][idx]
            cdata = np.mean(cdata, 1)
            kwargs['facecolordata'] = cdata
        return args, kwargs

    def plot_solid(self, *args, **kwargs):
        '''
        plot_solid(v)  or plot_solid(v, idx)

        v [element_index, points_in_element, xyz]

        or

        v [vertex_index, xyz]
        idx = [element_idx, point_in_element]

        kwargs: normals : normal vectors
        '''
        shade = kwargs.pop('shade', 'linear')
        if shade == 'flat':
            args, kwargs = self.prep_flat_shading_data(args, kwargs)

        if len(args) == 1:
            v = args[0]
            vv = v.reshape(-1, v.shape[-1])  # vertex
            nv = len(v[:, :, 2].flatten())
            idxset = np.arange(nv, dtype=int).reshape(v.shape[0], v.shape[1])
            nverts = v.shape[0] * v.shape[1]
            ncounts = v.shape[1]
            nele = v.shape[0]
        else:
            v = args[0]   # vertex
            vv = v
            idxset = np.array(args[1], dtype=int, copy=None)
            # element index (element_idx, point_in_element)
            nverts = v.shape[0]
            ncounts = idxset.shape[1]
            nele = idxset.shape[0]

        norms = kwargs.pop('normals', None)

        w = np.zeros((nverts))  # weight
        if norms is None:
            norms = np.zeros((nverts, 3), dtype=np.float32)  # weight
            if idxset.shape[1] > 2:
                xyz = vv[idxset[:, :3]].astype(float, copy=False)
                p0 = xyz[:, 0, :] - xyz[:, 1, :]
                p1 = xyz[:, 0, :] - xyz[:, 2, :]
                n1a = np.cross(p0, p1)
                da = np.atleast_2d(np.linalg.norm(n1a, axis=1)).transpose()
                da[da == 0.0] = 1.
                n1a = -n1a / da
            else:
                da = np.zeros(idxset.shape[0])
                n1a = np.zeros((nverts, 3), dtype=np.float32)  # weight
                n1a[:, 2] = 1.0

            if len(args) == 1:
                if da[0] == 0.:
                    norms[:, 2] = 1  # all [0. 0. 1]
                else:
                    for k in range(idxset.shape[1]):
                        norms[idxset[:, k], :] = n1a
            elif idxset.shape[-1] < 3:
                norms = n1a
            else:
                # print("doing this")
                # print(time.perf_counter())
                data = np.ones(idxset.flatten().shape[0])
                jj = np.tile(np.arange(idxset.shape[0]), idxset.shape[-1])
                ii = idxset.transpose().flatten()
                table = coo_matrix((data, (ii, jj)),
                                   shape=(nverts, idxset.shape[0]))
                csr = table.tocsr()
                indptr = csr.indptr
                indices = csr.indices
                data = csr.data
                '''
                for i in range(csr.shape[0]):
                    nn = n1a[indices[indptr[i]:indptr[i+1]]]
                    if len(nn) != 0.0:
                        sign = np.sign(np.sum(nn*nn[0], 1))
                        data[indices[indptr[i]:indptr[i+1]]] = sign
                    else:
                        pass
                        #norms[i, :] = [1,0,0]
                '''
                norms = table.dot(n1a)
                # print(time.perf_counter())
            nn = np.atleast_2d(np.linalg.norm(norms, axis=1)).transpose()
            nn[nn == 0.0] = 1.
            norms = norms / nn

        kwargs['gl_3dpath'] = [v[..., 0].flatten(),
                               v[..., 1].flatten(),
                               v[..., 2].flatten(),
                               norms, idxset]
        edge_idx = kwargs.pop('edge_idx', None)
        if edge_idx is not None:
            kwargs['gl_edge_idx'] = edge_idx

        from .art3d_gl import Poly3DCollectionGL

        if len(args) == 1:
            a = Poly3DCollectionGL(v[:2, ...], **kwargs)
        else:
            a = Poly3DCollectionGL(v[idxset[:2, ...]], **kwargs)

        # For GL aritsts, it is not necesasry to put in collections??
        # Axes3D.add_collection3d(self, a)
        Axes3D.add_artist(self, a)
        a.do_stencil_test = False

        return a

    def rotate_view_90deg(self, flip=False):
        self._vvec = self._vvec / np.sqrt(np.sum(self._vvec**2))
        if flip:
            M = rotation_mat(self._vvec, np.pi / 2)
        else:
            M = rotation_mat(self._vvec, -np.pi / 2)
        self._upvec = np.dot(M, self._upvec)

    def get_proj2(self):
        '''
        based on mplot3d::get_proj()
        it exposes matries used to compose projection matrix,
        and supports orthogonal projection.
        '''
        relev, razim = np.pi * self.elev / 180, np.pi * self.azim / 180

        xmin, xmax = self.get_xlim3d()
        ymin, ymax = self.get_ylim3d()
        zmin, zmax = self.get_zlim3d()

        # transform to uniform world coordinates 0-1.0,0-1.0,0-1.0
        worldM = world_transformation(xmin, xmax, ymin, ymax, zmin, zmax,
                                      view_scale=self._gl_scale)

        # look into the middle of the new coordinates
        R = np.array([0.5, 0.5, 0.5])

        xp = R[0] + np.cos(razim) * np.cos(relev) * self.dist
        yp = R[1] + np.sin(razim) * np.cos(relev) * self.dist
        zp = R[2] + np.sin(relev) * self.dist
        E = np.array((xp, yp, zp))

        # self.eye = E
        self._vvec = R - E
        self._vvec = self._vvec / np.sqrt(np.sum(self._vvec**2))

        if abs(relev) > np.pi / 2:
            V = np.array((0, 0, -1))
        else:
            V = np.array((0, 0, 1))
        V = self._upvec
        viewM = view_transformation(E, R, V)

        zfront, zback = -self.dist, self.dist
        # zfront, zback = self.dist-1, self.dist+1

        if self._use_frustum:
            perspM = persp_transformation(zfront, zback)
        else:
            a = (zfront + zback) / (zfront - zback)
            b = -2 * (zfront * zback) / (zfront - zback)
            perspM = np.array([[1, 0, 0, 0],
                               [0, 1, 0, 0],
                               [0, 0, b, a],
                               [0, 0, 0, self.dist]
                               # [0, 0, -1/10000., self.dist]
                               ])

        if self._ignore_screen_aspect_ratio:
            M = np.array([[1, 0, 0, 0], [0, 1, 0, 0],
                          [0, 0, 1, 0], [0, 0, 0, 1]])
        else:
            bb = self.get_window_extent()
            r = abs((bb.x1 - bb.x0) / (bb.y1 - bb.y0))
            if r >= 1:
                M = np.array([[1. / r, 0, 0, 0], [0, 1, 0, 0],
                              [0, 0, 1, 0], [0, 0, 0, 1]])
            else:
                M = np.array([[1, 0, 0, 0], [0, r, 0, 0],
                              [0, 0, 1, 0], [0, 0, 0, 1]])
        self._matrix_cache_extra = M

        perspM = np.dot(M, perspM)

        M0 = np.dot(viewM, worldM)
        M = np.dot(perspM, M0)

        self._matrix_cache_combined = M
        # perspM is used to draw Axes,,,
        return worldM, viewM, perspM, E, R, V, self.dist

    def get_proj(self):
        void = self.get_proj2()
        return self._matrix_cache_combined

    def set_lighting(self, *args, **kwargs):
        if len(args) != 0:
            kwargs = args[0]
        for k in kwargs:
            if k in self._lighting:
                self._lighting[k] = kwargs[k]

    def get_lighting(self):
        return self._lighting

    def show_3d_axes(self, value):
        self._show_3d_axes = value

    def _find_3d_loc_for_axis(self):

        minx, maxx, miny, maxy, minz, maxz = self.get_w_lims()
        vmax = np.array((maxx, maxy, maxz))
        vmin = np.array((minx, miny, minz))
        vrange = vmax - vmin
        vmid = (vmax + vmin) / 2.0

        from ifigure.utils.geom import transform_point
        x0s, y0s = transform_point(
            self.transAxes, 0.1, 0.1)
        x0, y0 = transform_point(
            self.transData.inverted(), x0s, y0s)

        x1s, y1s = transform_point(
            self.transAxes, 0.5, 0.5)
        x1, y1 = transform_point(
            self.transData.inverted(), x1s, y1s)

        # print("position here", x0, y0, x1, y1)
        # dd is the screen space move I want to make
        dd = np.array([-x1 + x0, -y1 + y0])

        center_move = self._screen_move_to_3d_move(dd)
        base = vmid + center_move

        dds = np.array([-x0s + x1s, -y0s + y1s])
        return base, vrange, dds

    def remove_3daxes_artists(self):
        if self._3d_axes_icon is None:
            return
        self._3d_axes_icon[0]().remove()
        self._3d_axes_icon[1]().remove()
        self._3d_axes_icon[2]().remove()
        self._3d_axes_icon[3]().remove()
        self._3d_axes_icon[4]().remove()
        self._3d_axes_icon[5]().remove()
        self._3d_axes_icon = None

    def draw_3d_axes(self):
        M = self._matrix_cache_combined
        # M = self.get_proj()

        def ptf2(x, dx):
            # xvec = np.dot(M, np.array([x[0], x[1], x[2], 1]))
            # xx1 = xvec[:2]/xvec[3]
            xvec = np.dot(M, np.array(
                [x[0] + dx[0], x[1] + dx[1], x[2] + dx[2], 1]))
            xx2 = xvec[:2] / xvec[3]
            return xx2

        if self._3d_axes_icon is None:
            base, lbase, dd = self._find_3d_loc_for_axis()

            # 35 ia an arbiray limit to keep the arrow from getting too big.
            scale = 0.1 if dd[0] * 0.1 < 35 else 35. / dd[0]

            vecx = np.array([1, 0, 0.])
            vecy = np.array([0, 1, 0.])
            vecz = np.array([0, 0, 1.])

            pt1 = ptf2(base, vecx * 2 * lbase[0] * scale)
            v = np.vstack([arrow3d(base,
                                   0.05 / lbase[0],
                                   0.25 / lbase[0],
                                   vecx,
                                   lbase[0] * scale,
                                   lbase[0] * scale * 0.8,
                                   m=13,
                                   d1d2=(vecy * lbase[1],
                                         vecz * lbase[2]))])

            kwargs = {'facecolor': 'r', 'edgecolor': 'r', 'linewidth': 0}
            a1 = self.plot_solid(v, **kwargs)
            a1._gl_always_noclip = True
            a1._gl_isArrow = True
            a1._gl_repr_name = 'x arrow'
            a1._gl_isLast = True

            pt2 = ptf2(base, vecy * 2 * lbase[1] * scale)
            v = np.vstack([arrow3d(base,
                                   0.05 / lbase[1],
                                   0.25 / lbase[1],
                                   vecy,
                                   lbase[1] * scale,
                                   lbase[1] * scale * 0.8,
                                   m=13,
                                   d1d2=(vecx * lbase[0],
                                         vecz * lbase[2]))])
            kwargs = {'facecolor': 'g', 'edgecolor': 'g', 'linewidth': 0}
            a2 = self.plot_solid(v, **kwargs)
            a2._gl_always_noclip = True
            a2._gl_isArrow = True
            a2._gl_repr_name = 'y arrow'
            a2._gl_isLast = True

            pt3 = ptf2(base, vecz * 2 * lbase[2] * scale)
            v = np.vstack([arrow3d(base,
                                   0.05 / lbase[2],
                                   0.25 / lbase[2],
                                   vecz,
                                   lbase[2] * scale,
                                   lbase[2] * scale * 0.8,
                                   m=13,
                                   d1d2=(vecx * lbase[0],
                                         vecy * lbase[1]))])
            kwargs = {'facecolor': 'b', 'edgecolor': 'b', 'linewidth': 0}
            a3 = self.plot_solid(v, **kwargs)
            a3._gl_always_noclip = True
            a3._gl_isArrow = True
            a3._gl_repr_name = 'z arrow'
            a3._gl_isLast = True

            a4 = Text(pt1[0], pt1[1], 'X', color='r',
                      axes=self, figure=self.figure,
                      transform=self.transData,
                      verticalalignment='center',
                      horizontalalignment='center')

            a5 = Text(pt2[0], pt2[1], 'Y', color='g',
                      axes=self, figure=self.figure,
                      transform=self.transData,
                      verticalalignment='center',
                      horizontalalignment='center')

            a6 = Text(pt3[0], pt3[1], 'Z', color='b',
                      axes=self, figure=self.figure,
                      transform=self.transData,
                      verticalalignment='center',
                      horizontalalignment='center')

            # self.texts.append(a4)
            # self.texts.append(a5)
            # self.texts.append(a6)
            self.add_artist(a4)
            self.add_artist(a5)
            self.add_artist(a6)

            self._3d_axes_icon = [weakref.ref(a1),
                                  weakref.ref(a2),
                                  weakref.ref(a3),
                                  weakref.ref(a4),
                                  weakref.ref(a5),
                                  weakref.ref(a6)]
        else:
            p0, p1, pt = ptf(xvec)
            # self._3d_axes_icon[0]().set_xdata(p0)
            # self._3d_axes_icon[0]().set_ydata(p1)
            self._3d_axes_icon[3]().set_x(pt[0])
            self._3d_axes_icon[3]().set_y(pt[1])

            p0, p1, pt = ptf(yvec)
            # self._3d_axes_icon[1]().set_xdata(p0)
            # self._3d_axes_icon[1]().set_ydata(p1)
            self._3d_axes_icon[4]().set_x(pt[0])
            self._3d_axes_icon[4]().set_y(pt[1])

            p0, p1, pt = ptf(zvec)
            # self._3d_axes_icon[2]().set_xdata(p0)
            # self._3d_axes_icon[2]().set_ydata(p1)
            self._3d_axes_icon[5]().set_x(pt[0])
            self._3d_axes_icon[5]().set_y(pt[1])

    def get_gl_uniforms(self):
        glcanvas = get_glcanvas()
        return glcanvas.get_uniforms()

    def apply_aspect(self, position=None):
        '''
        after MPL 3.3, aspect ratio is force to be one for 3D.
        we want to void it since our transformation support different
        aspect ratio

        self._box_aspect = None is needed reproduce the old versions
        behavior for MPL 3.3
        '''
        self._box_aspect = None
        super(Axes3D, self).apply_aspect(position)

    @allow_rasterization
    def draw(self, renderer):
        #        if self._use_gl and isSupportedRenderer(renderer):
        gl_len = 0

        if isSupportedRenderer(renderer):
            self._matrix_cache = self.get_proj2()

            if self._show_3d_axes:
                self.remove_3daxes_artists()
                self.draw_3d_axes()
            else:
                self.remove_3daxes_artists()

            artists = []

            artists.extend(self.images)
            artists.extend(self.collections)
            artists.extend(self.patches)
            artists.extend(self.lines)
            artists.extend(self.artists)
            artists.extend(self.texts)

            if self._show_3d_axes:
                zorder = max([a.get_zorder() for a in artists])
                for a in self._3d_axes_icon[3:]:
                    a().set_zorder(zorder + 5)

                # adjust padding based on lable/ticklable sizes
                xmax = max(*([x.get_size()
                           for x in self.xaxis.get_majorticklabels()] + [0]))
                ymax = max(*([x.get_size()
                           for x in self.yaxis.get_majorticklabels()] + [0]))
                zmax = max(*([x.get_size()
                           for x in self.zaxis.get_majorticklabels()] + [0]))
                self.xaxis.labelpad = (xmax + self.xaxis.label.get_size())/2.0
                self.yaxis.labelpad = (ymax + self.yaxis.label.get_size())/2.0
                self.zaxis.labelpad = (zmax + self.zaxis.label.get_size())/2.0

            # zorder = max([a.get_zorder() for a in artists])
            # print(zorder)

            gl_obj = [a for a in artists if hasattr(a, 'is_gl')]
            for a in artists:
                if hasattr(a, 'figobj'):
                    a._gl_always_noclip = a.figobj.getp("noclip3d")

            gl_len = len(gl_obj)
            if gl_len > 0:
                glcanvas = get_glcanvas()
                if (glcanvas is not None and
                        glcanvas.init):

                    params = self._lighting.copy()
                    del params["clip_limit1"]
                    del params["clip_limit2"]
                    xmin, xmax = self.get_xlim3d()
                    ymin, ymax = self.get_ylim3d()
                    zmin, zmax = self.get_zlim3d()

                    cc = np.array(
                        [(xmin+xmax)/2.0, (ymin+ymax)/2.0, (zmin+zmax)/2.0])
                    l1 = self._lighting["clip_limit1"]
                    l2 = self._lighting["clip_limit2"].copy()

                    nn = l1*np.array([xmax-xmin, ymax-ymin, zmax-zmin])
                    nn2 = np.sqrt(np.sum(nn*nn))

                    if nn2 != 0:
                        l2[0] = -(l2[0] + np.sum(l1*cc))/nn2
                        l1 = nn/nn2

                    glcanvas.set_lighting(
                        clip_limit1=l1, clip_limit2=l2, **params)
#                   glcanvas.set_lighting(**self._lighting)

                    glcanvas._gl_scale = self._gl_scale
                else:
                    return
            renderer._num_globj = gl_len
            renderer._k_globj = 0
        else:
            assert False, "Unsupported Renderer"

        # axes3D seems to change frameon status....
        if callable(self.get_frame_on):
            frameon = self.get_frame_on()
            self.set_frame_on(False)
        else:
            pass  # matplotlib > 3.8

        if self._gl_scale != 1.0:
            # print(("gl_scale", self._gl_scale))
            xmin, xmax = self.get_xlim3d()
            ymin, ymax = self.get_ylim3d()
            zmin, zmax = self.get_zlim3d()
            dx = (xmax - xmin) / self._gl_scale / 2.
            dy = (ymax - ymin) / self._gl_scale / 2.
            dz = (zmax - zmin) / self._gl_scale / 2.
            xmid = (xmax + xmin) / 2.
            ymid = (ymax + ymin) / 2.
            zmid = (zmax + zmin) / 2.
            self.set_xlim3d([xmid - dx, xmid + dx])
            self.set_ylim3d([ymid - dy, ymid + dy])
            self.set_zlim3d([zmid - dz, zmid + dz])

        val = Axes3D.draw(self, renderer)

        if self._gl_scale != 1.0:
            self.set_xlim3d([xmin, xmax])
            self.set_ylim3d([ymin, ymax])
            self.set_zlim3d([zmin, zmax])

        if callable(self.set_frame_on):
            self.set_frame_on(frameon)
        else:
            pass  # matplotlib > 3.8

        return val

    def _screen_move_to_3d_move(self, delta_in_axesnorm):
        #
        # compute the linear transformation in 3D,
        # which causes the transformation given by dd (axes normal coords)
        #
        # the computed 3D transformation is normal to screen normal
        #
        minx, maxx, miny, maxy, minz, maxz = self.get_w_lims()

        vmax = np.array((maxx, maxy, maxz))
        vmin = np.array((minx, miny, minz))
        vrange = vmax - vmin
        vmid = (vmax + vmin) / 2.0
        midx, midy, midz = vmid

        M = self._matrix_cache_combined

        def vec2sc(vec):
            x = np.dot(M, (vec[0], vec[1], vec[2], 1))
            return x[0:2] / x[-1]

        xx = vec2sc((maxx, midy, midz,)) - vec2sc((minx, midy, midz,))
        yy = vec2sc((midx, maxy, midz,)) - vec2sc((midx, miny, midz,))
        zz = vec2sc((midx, midy, maxz,)) - vec2sc((midx, midy, minz,))

        from scipy.optimize import nnls
        from scipy.linalg import null_space
        A = np.vstack((xx, yy, zz)).transpose()

        x = (null_space(A)).flatten()
        nx = norm_vec(x)

        v1 = np.cross(nx, [1, 0, 0]) * vrange
        v2 = np.cross(nx, [0, 1, 0]) * vrange
        v3 = np.cross(nx, [0, 0, 1]) * vrange
        # print("v1, v2, v3", v1,  v2, v3)

        def screen_diff(vv):
            return (vec2sc((midx + vv[0], midy + vv[1], midz + vv[2])) -
                    vec2sc((midx - vv[0], midy - vv[1], midz - vv[2]))) / 2.0
        d1 = screen_diff(v1)
        d2 = screen_diff(v2)
        d3 = screen_diff(v3)

        c1 = abs(cross2d(norm_vec(d2), norm_vec(d3)))
        c2 = abs(cross2d(norm_vec(d1), norm_vec(d3)))
        c3 = abs(cross2d(norm_vec(d1), norm_vec(d2)))
        # print("c1, c2, c3", c1,  c2, c3)
        vv = [v1, v2, v3]
        dd = [d1, d2, d3]

        ii = np.argsort([c1, c2, c3])
        iii = ((1, 2), (0, 2), (0, 1))[ii[-1]]

        v1 = vv[iii[1]]
        v2 = vv[iii[0]]
        d1 = dd[iii[1]]
        d2 = dd[iii[0]]

        A = np.vstack((d1, d2)).transpose()

        sol = np.dot(np.linalg.inv(A), delta_in_axesnorm)
        center_move = (v1 * sol[0] + v2 * sol[1])

        return center_move

    def _screen_shift_3drot(self, Mold, Mnew):
        #
        # compute shift of screen for 3d zoom
        # 3d rotation is done around _gl_rot_center
        #
        minx, maxx, miny, maxy, minz, maxz = self.get_w_lims()

        vmax = np.array((maxx, maxy, maxz))
        vmin = np.array((minx, miny, minz))
        vrange = vmax - vmin
        vmid = (vmax + vmin) / 2.0
        midx, midy, midz = vmid

        def vec2sc_old(vec):
            x = np.dot(Mold, (vec[0], vec[1], vec[2], 1))
            return x[0:2] / x[-1]

        def vec2sc_new(vec):
            x = np.dot(Mnew, (vec[0], vec[1], vec[2], 1))
            return x[0:2] / x[-1]

        # print("rot_center", vec2sc_old(self._gl_rot_center), vec2sc_new(self._gl_rot_center))

        dd = vec2sc_new(self._gl_rot_center) - vec2sc_old(self._gl_rot_center)
        center_move = self._screen_move_to_3d_move(dd)

        minp = np.array([minx, miny, minz])
        maxp = np.array([maxx, maxy, maxz])

        r1 = minp + center_move
        r2 = maxp + center_move

        return (r1[0], r2[0]), (r1[1], r2[1]), (r1[2], r2[2])

    def set_gl_scale_accum(self, value):
        self._gl_scale_accum = value

    def get_gl_scale_accum(self):
        return self._gl_scale_accum



import wx
from ifigure.widgets.redirect_output import RedirectOutput


class Consol(wx.Panel):
    def __init__(self, parent, *args, **kargs):
        super(Consol, self).__init__(parent, *args, **kargs)
        self.log = wx.TextCtrl(
            self, -1, style=wx.TE_MULTILINE | wx.TE_READONLY)
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.log, 1, wx.EXPAND)
        self.SetSizer(sizer)



from __future__ import print_function
#
#  Name   :fig_fill.py
#
#          FigFill is the class to support patchcollections.
#          It supports Axes.fill, fill_between, and fill_betweenx.
#          A pathce attritube is changed simultaneously.
#
#          Matplotlib generates either Polygon or PolyCollection
#          FigFill absorbes the differecne of set_xxx/get_xxx of
#          these two.
#
#          Changes:
#             2014 02 11 properties are edited using mode=2 (not mode=1)
#
#
#
#
#  Author :
#         Syun'ichi Shiraiwa
#  E-mail :
#         shiraiwa@psfc.mit.edu
#
#  History :
#         2013 03    ver.0.1
#              04    ver.0.2 large rewrite to unifiy interface

from ifigure.mto.fig_obj import FigObj, mask_negative
from ifigure.mto.axis_user import XUser, YUser, ZUser
from ifigure.widgets.canvas.file_structure import *
import ifigure
import os
import ifigure.widgets.canvas.custom_picker as cpicker
import numpy as np
import weakref
import logging
from ifigure.utils.cbook import isiterable, ProcessKeywords, LoadImageFile
import matplotlib
from matplotlib.collections import PolyCollection
from matplotlib.patches import Polygon

from ifigure.utils.args_parser import ArgsParser

import ifigure.utils.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('FigFill')


def _arg_names(mpl_cmd):
    if mpl_cmd == 'fill':
        return ("x", "y")
    elif mpl_cmd == 'fill_between':
        return ("x", "y", "y2", "where")
    elif mpl_cmd == 'fill_betweenx':
        return ("y", "x", "x2", "where")


class FigFill(FigObj, XUser, YUser, ZUser):
    def __new__(cls, *args, **kywds):
        def set_hidden_vars(obj):
            obj._mpl_cmd = 'fill'
            obj._data_extent = None
            return obj

        if 'src' in kywds:
            obj = FigObj.__new__(cls, *args, **kywds)
            obj = set_hidden_vars(obj)
            return obj

        mpl_cmd, kywds = ProcessKeywords(kywds, 'mpl_command', 'fill')
        p = ArgsParser()

        if mpl_cmd == 'fill_betweenx':
            p.add_var('y', ['iter|nonstr', 'dynamic'])
            p.add_opt('x', None, ['iter|nonstr', 'dynamic'])
        else:
            p.add_opt('x', None, ['iter|nonstr', 'dynamic'])
            p.add_var('y', ['iter|nonstr', 'dynamic'])
        #if mpl_cmd == 'fill':
        #    p.add_opt('s', '', 'str')

        p.add_key('x2', None)
        p.add_key('y2', None)
        p.add_key('where', None)
        p.add_key('edgecolor', 'black')
        p.add_key('linewidth', 1.0)
        p.add_key('linestyle', 'solid')
        p.add_key('alpha', 1)
        p.add_key('facecolor', 'blue')
        p.add_key('zs', 0)

        p.set_ndconvert("x", "y", "x2", "y2")

        v, kywds, d, flag = p.process(*args, **kywds)

        if not flag:
            raise ValueError('Failed when processing argument')

        obj = FigObj.__new__(cls, *args, **kywds)
        obj = set_hidden_vars(obj)
        obj._mpl_cmd = mpl_cmd
        if v["x"] is None:
            v["x"] = np.arange(v["y"].shape[-1])
        for name in v:
            obj.setvar(name, v[name])
        obj.setvar("kywds", kywds)
        return obj

    def __init__(self, *args,  **kywds):
        XUser.__init__(self)
        YUser.__init__(self)
        ZUser.__init__(self)

        args = []
        if 'src' not in kywds:
            kywds = self.getvar("kywds")
        super(FigFill, self).__init__(*args, **kywds)

    @classmethod
    def isFigFill(self):
        return True

    @classmethod
    def can_have_child(self, child=None):
        return True

    @classmethod
    def get_namebase(self):
        return 'fill'

    @classmethod
    def property_in_file(self):
        return []

    @classmethod
    def property_in_palette(self):
        return (["path", "patch"],
                [["edgecolor_2", "linewidth_2",
                  "plinestyle_2", "alpha_2"],
                 ["facecolor_2"]])

    @classmethod
    def attr_in_file(self):
        return (["edgecolor", "facecolor", "linewidth", "linestyle",
                 "alpha"]+super(FigFill, self).attr_in_file())

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = LoadImageFile(path, 'fill.png')
        return [idx1]

    def set_parent(self, parent):
        XUser.unset_ax(self)
        YUser.unset_ay(self)
        ZUser.unset_az(self)

        super(FigFill, self).set_parent(parent)
        XUser.get_xaxisparam(self)
        YUser.get_yaxisparam(self)
        ZUser.get_zaxisparam(self)

    def args2var(self):
        names0 = self.attr_in_file()

        if self._mpl_cmd == 'fill':
            names = ["x", "y", "s"]
            use_np = [True, True, False]
        elif self._mpl_cmd == 'fill_between':
            names = ["x", "y", "y2", "where"]
            use_np = [True, True, True, True]
        elif self._mpl_cmd == 'fill_betweenx':
            names = ["x", "y", "x2", "where"]
            use_np = [True, True, True, True]

        names = names + names0
        use_np = use_np + [False]*len(names0)
        values = self.put_args2var(names, use_np)

        x = values[0]
        y = values[1]
        if x is None:
            x = np.arange((y.shape)[-1])
            self.setp("x", x)
        if x is None:
            return False
        if y is None:
            return False

        if self._mpl_cmd == 'fill_between':
            if self.getp('y2') is None:
                self.setp('y2', np.zeros(self.getp('y').size))
        elif self._mpl_cmd == 'fill_betweenx':
            if self.getp('x2') is None:
                self.setp('x2', np.zeros(self.getp('x').size))

        return True

    def generate_artist(self):
        # this method generate artist
        # if artist does exist, update artist
        # based on the information specifically
        # managed by fig_obj tree. Any property
        # internally managed by matplotlib
        # does not change

        container = self.get_container()

        if self.isempty() is False:
            return
        s = ""
        if self._mpl_cmd == 'fill':
            x, y = self._eval_xy()
        elif self._mpl_cmd == 'fill_between':
            x, y, y2, where = self._eval_xy()
        elif self._mpl_cmd == 'fill_betweenx':
            y, x, x2, where = self._eval_xy()

        kywds = self._var["kywds"].copy()

        kywds["alpha"] = self.getp('alpha')
        kywds["edgecolor"] = self.getp('edgecolor')
        kywds["linewidth"] = self.getp('linewidth')
        kywds["facecolor"] = self.getp('facecolor')
        kywds["linestyle"] = self.getp('linestyle')

        try:
            if self.get_figaxes().get_3d():
                kywds['zs'] = self.getvar('zs')
        except:
            pass
        try:
            if self._mpl_cmd == 'fill':
                self._artists = container.fill(x, y, **kywds)
            elif self._mpl_cmd == 'fill_between':
                kywds["y2"] = y2
                kywds["where"] = where
                self._artists = [container.fill_between(x, y, **kywds)]
            elif self._mpl_cmd == 'fill_betweenx':
                kywds["x2"] = x2
                kywds["where"] = where
                self._artists = [container.fill_betweenx(y, x, **kywds)]
        except:
            dprint1('Failed to generate artilst')
            import traceback
            print(traceback.format_exc())
            self._artists = []
            return
            pass

#        self.set_data_extent()    # set data extent
        dprint2(self._artists)
        lp = self.getp("loaded_property")
        if lp is not None:
            # do apply saved property...
            for i in range(len(lp)):
                if i < len(self._artists):
                    self.set_artist_property(self._artists[i], lp[i])
            self.delp("loaded_property")

        for artist in self._artists:
            artist.figobj = self
            artist.figobj_hl = []
            artist.set_zorder(self.getp('zorder'))

    def del_artist(self, artist=None, delall=False):
        if delall:
            artistlist = self._artists
        else:
            artistlist = artist

        self.store_loaded_property()

        if len(artistlist) != 0:
            self.highlight_artist(False, artistlist)
            container = self.get_container()
            for a in artistlist:
                try:
                    a.remove()
                except:
                    logging.exception("fig_fill:: highlight_aritst() failed")

        super(FigFill, self).del_artist(artistlist)

    def highlight_artist(self, val, artist=None):
        if artist is None:
            alist = self._artists
        else:
            alist = artist
        container = self.get_container()
        from ifigure.matplotlib_mod.art3d_gl import Polygon3DGL, \
            Poly3DCollectionGL
        if val == True:
            for a in alist:
                if isinstance(a, Polygon3DGL):
                    pass
                elif isinstance(a, Poly3DCollectionGL):
                    hl = a.make_hl_artist(container)
                    for item in hl:
                        a.figobj_hl.append(item)

                elif hasattr(a, 'get_xy'):
                    xy = a.get_xy()
                    hl = container.plot(xy[:, 0], xy[:, 1], marker='s',
                                        color='k', linestyle='None',
                                        markerfacecolor='none',
                                        markeredgewidth=0.5,
                                        scalex=False, scaley=False)
                    for item in hl:
                        a.figobj_hl.append(item)
                else:
                    x = np.concatenate([p.vertices[:, 0]
                                        for p in a.get_paths()])
                    y = np.concatenate([p.vertices[:, 1]
                                        for p in a.get_paths()])
                    hl = container.plot(x, y, marker='s',
                                        color='k', linestyle='None',
                                        markerfacecolor='none',
                                        markeredgewidth=0.5,
                                        scalex=False, scaley=False)
                    for item in hl:
                        a.figobj_hl.append(item)
        else:
            for a in alist:
                for hl in a.figobj_hl:
                    hl.remove()
                a.figobj_hl = []
#
#   def hit_test
#

    def picker_a(self, artist, evt):
        axes = artist.axes
        if axes is None:
            return False, {}

        hit, extra = artist.contains(evt)

        if hit:
            return True, {'child_artist': artist}
        else:
            #            print('double check for picker')
            if hasattr(artist, 'get_verts'):
                # this is for Polygon object
                figure = self.get_figpage()._artists[0]
                v = artist.get_verts()
                if len(v) == 0:
                    return False, {}
                hl = matplotlib.lines.Line2D(v[:, 0], v[:, 1], marker='s',
                                             color='k', linestyle='None',
                                             markerfacecolor='none',
                                             figure=figure)
                figure.lines.extend([hl])
                hit, extra = hl.contains(evt)
                figure.lines.remove(hl)
                if hit:
                    return True, {'child_artist': artist}
            return False, {}

    def picker_a0(self, artist, evt):
        hit, extra = self.picker_a(artist, evt)
        if hit:
            return hit, extra, 'area', 3
        else:
            return False, {}, None, 0

#
#  Setter/Getter
#
    def _fill_setter(self, prop, value):
        from ifigure.utils.cbook import isiterable
        for a in self._artists:
            if isinstance(a, PolyCollection):
                m0 = getattr(a, 'get_'+prop)
                m = getattr(a, 'set_'+prop)
                if isiterable(m0()):
                    m([value]*len(m0()))
                else:
                    m(value)
            elif isinstance(a, Polygon):
                m = getattr(a, 'set_'+prop)
                m(value)
            else:
                dprint1('FigFill::Unknown artist type')

    def set_facecolor(self, value, a):
        self.setp('facecolor', value)
        self._fill_setter('facecolor', value)

    def get_facecolor(self, value):
        return self.getp('facecolor')

    def set_edgecolor(self, value, a):
        self.setp('edgecolor', value)
        self._fill_setter('edgecolor', value)

    def get_edgecolor(self, value):
        return self.getp('edgecolor')

    def set_linewidth(self, value, a):
        self.setp('linewidth', value)
        self._fill_setter('linewidth', value)

    def get_linewidth(self, value):
        return self.getp('linewidth')

    def set_linestyle(self, value, a):
        self.setp('linestyle', value)
        self._fill_setter('linestyle', value)

    def get_linestyle(self, value):
        return self.getp('linestyle')

    def set_alpha(self, value, a):
        self.setp('alpha', value)
        self._fill_setter('alpha', value)

    def get_alpha(self, value):
        return self.getp('alpha')

#
#   data extent
#
    def get_data_extent(self):
        if self._data_extent is not None:
            return self._data_extent
        if self._mpl_cmd == 'fill':
            x, y = self._eval_xy()
            if x is None:
                self._data_extent = [None]*4
            else:
                self._data_extent = [np.min(x), np.max(x),
                                     np.min(y), np.max(y)]
        elif self._mpl_cmd == 'fill_between':
            x, y, y2, where = self._eval_xy()
            if x is None:
                self._data_extent = [None]*4
            else:
                self._data_extent = [np.min(x), np.max(x),
                                     min([np.min(y), np.min(y2)]),
                                     max([np.max(y), np.max(y2)])]
        elif self._mpl_cmd == 'fill_betweenx':
            y, x, x2, where = self._eval_xy()
            if x is None:
                self._data_extent = [None]*4
            else:
                self._data_extent = [min([np.min(x), np.min(x2)]),
                                     max([np.max(x), np.max(x2)]),
                                     np.min(y), np.max(y)]

        return self._data_extent
#
#   range
#

    def get_xrange(self, xrange=[None, None], scale='linear'):
        if self._mpl_cmd == 'fill':
            x, y = self._eval_xy()
            if x is None:
                return
            if scale == 'log':
                x = mask_negative(x)
            return self._update_range(xrange, [np.nanmin(x), np.nanmax(x)])
        elif self._mpl_cmd == 'fill_between':
            x, y, y2, where = self._eval_xy()
            if x is None:
                return
            if scale == 'log':
                x = mask_negative(x)
            return self._update_range(xrange, [np.nanmin(x), np.nanmax(x)])
        elif self._mpl_cmd == 'fill_betweenx':
            y, x, x2, where = self._eval_xy()
            if x is None:
                return
            if x2 is None:
                return
            if scale == 'log':
                x = mask_negative(x)
                x2 = mask_negative(x2)
            return self._update_range(xrange, [np.nanmin([np.nanmin(x), np.nanmin(x2)]),
                                               np.nanmax([np.nanmax(x), np.nanmax(x2)])])

    def get_yrange(self, yrange=[None, None], xrange=[None, None], scale='linear'):
        y0 = yrange[0]
        y1 = yrange[1]
        y0n = None
        y1n = None
        if (xrange[0] is not None and
                xrange[1] is not None):
            if self._mpl_cmd == 'fill':
                x, y = self._eval_xy()
                ym = np.ma.masked_array(y)
                ym[x < xrange[0]] = np.ma.masked
                ym[x > xrange[1]] = np.ma.masked
                if scale == 'log':
                    ym[x < 0] = np.ma.masked
                return self._update_range(yrange, [np.nanmin(ym), np.nanmax(ym)])
            elif self._mpl_cmd == 'fill_between':
                x, y, y2, where = self._eval_xy()
                ym = np.ma.masked_array(y)
                y2m = np.ma.masked_array(y2)
#                print y, y2
                ym[x < xrange[0]] = np.ma.masked
                ym[x > xrange[1]] = np.ma.masked
                y2m[x < xrange[0]] = np.ma.masked
                y2m[x > xrange[1]] = np.ma.masked
                if scale == 'log':
                    ym[x < 0] = np.ma.masked
                    y2m[x < 0] = np.ma.masked
                return self._update_range(yrange, [np.nanmin([np.nanmin(ym),  np.nanmin(y2m)]),
                                                   np.nanmax([np.nanmax(ym), np.nanmax(y2m)])])
            elif self._mpl_cmd == 'fill_betweenx':
                y, x, x2, where = self._eval_xy()
                ym = np.ma.masked_array(y)
                ym[x < xrange[0]] = np.ma.masked
                ym[x > xrange[1]] = np.ma.masked
                if scale == 'log':
                    ym[x < 0] = np.ma.masked
                return self._update_range(yrange, [np.nanmin(ym), np.nanmax(ym)])
        return yrange

    def get_zrange(self, zrange=[None, None],
                   xrange=[None, None],
                   yrange=[None, None], scale='linear'):
        return self._update_range(zrange, [self.getvar('zs'), self.getvar('zs')])
#
#   export
#

    def get_export_val(self, a):
        v = {"xdata": self.getvar('x'),
             "ydata": self.getvar('y')}
        if self.getvar('x2') is not None:
            v["x2data"] = self.getvar('x2')
        if self.getvar('y2') is not None:
            v["y2data"] = self.getvar('y2')
        return v

#
#   save/load
#
    def save_data2(self, data=None):
        def check(obj, name):
            if not isinstance(obj.getp(name), np.ndarray):
                return False
            if not isinstance(obj.getvar(name), np.ndarray):
                return False
            return obj.getp(name) is obj.getvar(name)

        if data is None:
            data = {}
        var = {'x': check(self, 'x'),
               'y': check(self, 'y'),
               'x2': check(self, 'x2'),
               'y2': check(self, 'y2'),
               'where': check(self, 'where'),
               's': check(self, 's'), }

        if var["x"] is not True:
            var["xdata"] = self.getp("x")
        if var["y"] is not True:
            var["ydata"] = self.getp("y")
        if var["x2"] is not True:
            var["x2data"] = self.getp("x2")
        if var["y2"] is not True:
            var["y2data"] = self.getp("y2")
        if var["where"] is not True:
            var["wheredata"] = self.getp("where")
        if var["s"] is not True:
            var["sdata"] = self.getp("s")
        var["mpl_cmd"] = self._mpl_cmd

        data['FigFill'] = (1, var)
        return super(FigFill, self).save_data2(data)

    def load_data2(self, data):
        d = data['FigFill']
        super(FigFill, self).load_data2(data)
        var = d[1]

        self._mpl_cmd = var["mpl_cmd"]
        names = ["x", "y", "x2", "y2", "s", "where"]
        for name in names:
            if var[name]:
                self.setp(name,
                          self.getvar(name))
            else:
                self.setp(name, var[name+'data'])

    def _eval_xy(self):
        names = _arg_names(self._mpl_cmd)
        if self.getp('use_var'):
            success = self.handle_use_var()
            if not success:
                return [None]*len(names)
        return self.getp(names)



from .backend_wxagg_gl import mixin_gl_renderer
from matplotlib.backends.backend_pdf import RendererPdf, PdfFile, Name, Op, pdfRepr
from matplotlib.backends.backend_ps import RendererPS
from matplotlib.backends.backend_svg import RendererSVG


def isPDFrenderer(renderer):
    return (hasattr(renderer, '_vector_renderer') and
            isinstance(renderer._vector_renderer, RendererPdf))


def isPSrenderer(renderer):
    return (hasattr(renderer, '_vector_renderer') and
            isinstance(renderer._vector_renderer, RendererPS))


def isSVGrenderer(renderer):
    return (hasattr(renderer, '_vector_renderer') and
            isinstance(renderer._vector_renderer, RendererSVG))


def isSupportedRenderer(renderer):
    if hasattr(renderer, '_gl_renderer'):
        return True
    elif isPDFrenderer(renderer):
        mixin_gl_renderer(renderer)
        return True
    elif isPSrenderer(renderer):
        mixin_gl_renderer(renderer)
        return True
    elif isSVGrenderer(renderer):
        mixin_gl_renderer(renderer)
        renderer.gl_svg_rescale = True
        return True
    return False



'''
mini frame with window list

   support accelerator so that F1, F2 key works
   on miniframe
'''

import wx


class WithWindowList_MixIn(object):
    def __init__(self):
        super(WithWindowList_MixIn, self).__init__()

        frame = self.GetParent()
        self._atable = []
        self._atable.append((wx.ACCEL_NORMAL,  wx.WXK_F2, wx.ID_BACKWARD))
        self._atable.append((wx.ACCEL_NORMAL,  wx.WXK_F1, wx.ID_FORWARD))
        atable = wx.AcceleratorTable(self._atable)
        self.SetAcceleratorTable(atable)
        self.Bind(wx.EVT_MENU, lambda evt: frame.ProcessEvent(evt))


class MiniFrameWithWindowList(wx.MiniFrame, WithWindowList_MixIn):
    def __init__(self, *args, **kargs):
        super(MiniFrameWithWindowList, self).__init__(*args, **kargs)
        WithWindowList_MixIn.__init__(self)
        '''
        frame = self.GetParent()
        self._atable = []
        self._atable.append((wx.ACCEL_NORMAL,  wx.WXK_F2, wx.ID_BACKWARD))
        self._atable.append((wx.ACCEL_NORMAL,  wx.WXK_F1, wx.ID_FORWARD))   
        atable = wx.AcceleratorTable(self._atable)
        self.SetAcceleratorTable(atable)
        self.Bind(wx.EVT_MENU, lambda evt: frame.ProcessEvent(evt))
        '''


class DialogWithWindowList(wx.Dialog, WithWindowList_MixIn):
    def __init__(self, *args, **kargs):
        style = kargs.pop('style', wx.DEFAULT_DIALOG_STYLE)
        kargs['style'] = style
        super(DialogWithWindowList, self).__init__(*args, **kargs)
        WithWindowList_MixIn.__init__(self)

        '''
        frame = self.GetParent()
        self._atable = []
        self._atable.append((wx.ACCEL_NORMAL,  wx.WXK_F2, wx.ID_BACKWARD))
        self._atable.append((wx.ACCEL_NORMAL,  wx.WXK_F1, wx.ID_FORWARD))   
        atable = wx.AcceleratorTable(self._atable)
        self.SetAcceleratorTable(atable)
        self.Bind(wx.EVT_MENU, lambda evt: frame.ProcessEvent(evt))
        '''



from .gl_compound import GLCompound
import ifigure.utils.debug as debug
from ifigure.mto.fig_obj import FigObj, mask_negative
from ifigure.mto.axis_user import XUser, YUser, ZUser, CUser
from ifigure.widgets.canvas.file_structure import *
import ifigure
import os
import ifigure.widgets.canvas.custom_picker as cpicker
import numpy as np
import weakref
import logging
from ifigure.utils.cbook import ProcessKeywords, LoadImageFile, isdynamic
from matplotlib.tri import Triangulation
from ifigure.matplotlib_mod.triplot_mod import triplot
from ifigure.utils.args_parser import ArgsParser
from matplotlib import cm
from matplotlib.colors import Normalize, colorConverter, LightSource
from matplotlib.cm import ScalarMappable
from matplotlib.colors import ColorConverter
cc = ColorConverter()
dprint1, dprint2, dprint3 = debug.init_dprints('FigSolid')

#from mpl_toolkits.axes_grid1.inset_locator import inset_axes


class FigSolid(GLCompound, FigObj, XUser, YUser, ZUser, CUser):
    def __new__(cls, *args, **kywds):
        def set_hidden_vars(obj):
            obj._objs = []  # for debug....
            obj._data_extent = None
            return obj

        if 'src' in kywds:
            obj = FigObj.__new__(cls, *args, **kywds)
            obj = set_hidden_vars(obj)
            return obj

        p = ArgsParser()
        p.add_var('v', ['iter|nonstr', 'dynamic'])
        p.add_opt('idxset', None, ['numbers|nonstr', 'dynamic'])

        #p.add_key('cmap', None)

        if 'cz' in kywds and kywds['cz']:
            def_alpha = None
            def_ec = None
            def_fc = None
            def_lw = 0.0
        else:
            def_alpha = 1.0
            def_ec = (0, 0, 0, 1)
            def_fc = (0, 0, 1, 1)
            def_lw = 1.0

        # this prevent from passing linewidth to aritist
        def_lw = kywds.pop('linewidth', def_lw)
        p.add_key('linewidths', def_lw)

        p.add_key('alpha', def_alpha)
        p.add_key('facecolor', def_fc)
        p.add_key('edgecolor', def_ec)
        p.add_key('normals', None)
        p.add_key('cz', False, 'bool')
        p.add_key('cdata', None)
        p.add_key('shade', 'linear')
        p.add_key('array_idx', None)
        p.add_key('edge_idx', None)
        p.add_key('use_pointfill', False)
        p.add_key('draw_last', False)

        v, kywds, d, flag = p.process(*args, **kywds)
        if not flag:
            raise ValueError('Failed when processing argument')

        obj = FigObj.__new__(cls, *args, **kywds)
        obj = set_hidden_vars(obj)

        for name in v:
            obj.setvar(name, v[name])
        obj.setvar("kywds", kywds)
        return obj

    def __init__(self, *args,  **kywds):
        XUser.__init__(self)
        YUser.__init__(self)
        ZUser.__init__(self)
        CUser.__init__(self)
        GLCompound.__init__(self)

        args = []
        if 'src' not in kywds:
            kywds = self.getvar("kywds")
        super(FigSolid, self).__init__(*args, **kywds)

    @classmethod
    def get_namebase(self):
        return 'solid'

    @classmethod
    def property_in_palette(self):
        return ["facecolor_2", "edgecolor_2", "linewidthz",
                "solid_shade", "alpha_2", "noclip3d"]

    @classmethod
    def attr_in_file(self):
        return (["alpha", "facecolor",
                 "edgecolor", "linewidth"] +
                super(FigSolid, self).attr_in_file())

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = LoadImageFile(path, 'surf.png')
        return [idx1]

    def set_parent(self, parent):
        XUser.unset_ax(self)
        YUser.unset_ay(self)
        ZUser.unset_az(self)
        CUser.unset_ac(self)
        super(FigSolid, self).set_parent(parent)
        XUser.get_xaxisparam(self)
        YUser.get_yaxisparam(self)
        ZUser.get_zaxisparam(self)
        CUser.get_caxisparam(self)

    def args2var(self):
        ret = self._args2var()
        return ret

    def _args2var(self):

        names0 = self.attr_in_file()
        names = ["v"]
        use_np = [True]
        names = names + names0
        use_np = use_np + [False]*len(names0)

        values = self.put_args2var(names, use_np)
        v = values[0]
        # need to do it manually since name is slightly different
        self.setp('linewidth', self.getvar('linewidths'))
        if v is None:
            return False
        self.setp('v', v)

        return True

    def generate_artist(self, coarse=False):
        if (self.isempty() is False
                and not coarse):
            return

        axes = self.get_figaxes()
        if not axes.get_3d():
            return

        container = self.get_container()

        v = self._eval_v()
        idxset = self.getvar('idxset')
        # use_var should be false if evaluation is
        # okey.
        if self.getp('use_var'):
            return

        lp = self.getp("loaded_property")
        if lp is not None:
            self.delp("loaded_property")

        norms = self.getvar('normals')
        '''
        if norms is None:
            norms = []
            for xyz in v:
                if xyz.shape[0] > 2:
                    p0, p1, p2 = [xyz[k,:3] for k in range(3)]
                    n1 = np.cross(p0-p1, p1-p2)
                    d = np.sqrt(np.sum(n1**2))
                else:
                    d = 0
                if d == 0:
                    norms.append([0,0,1]*xyz.shape[0])
                else:
                    norms.extend([-n1/d]*xyz.shape[0])
            norms = np.hstack(norms).astype(np.float32).reshape(-1,3)
        '''
        kywds = self._var["kywds"].copy()
        kywds['normals'] = norms
        # if self.getp('alpha') is not None else 1
        kywds['alpha'] = self.getp('alpha')
        kywds['array_idx'] = self.getvar('array_idx')
        kywds['edge_idx'] = self.getvar('edge_idx')
        fc = self.getp('facecolor')
        if isinstance(fc, str):
            fc = cc.to_rgba(fc)
        if fc is None:
            fc = [0, 0, 0, 0]
        else:
            fc = list(fc)
            if self.getp('alpha') is not None:
                fc[3] = self.getp('alpha')
        ec = self.getp('edgecolor')
        if isinstance(ec, str):
            ec = cc.to_rgba(ec)
        if ec is None:
            ec = [0, 0, 0, 0]
        else:
            ec = list(ec)
            if self.getp('alpha') is not None:
                ec[3] = self.getp('alpha')
        if self.getvar('cz'):
            kywds['cz'] = self.getvar('cz')
            if self.getvar('cdata') is not None:
                cdata = self.getvar('cdata')
            else:
                cdata = v[..., -1]
            if idxset is None:
                if self.getvar('shade') != 'linear':
                    kywds['facecolordata'] = np.mean(cdata, -1).real
                else:
                    kywds['facecolordata'] = cdata.real
            else:
                kywds['facecolordata'] = cdata.real
        else:
            kywds['facecolor'] = (fc,)
        kywds['edgecolor'] = (ec,)
        kywds['linewidths'] = 0.0 if self.getp(
            'linewidth') is None else self.getp('linewidth')
        kywds['shade'] = self.getvar('shade')
        kywds['use_pointfill'] = self.getvar('use_pointfill')

        args = (v[..., :3],) if idxset is None else (v[..., :3], idxset)
        self._artists = [container.plot_solid(*args, **kywds)]

        if self.getvar('draw_last'):
            self._artists[0]._gl_isLast = True

        self.set_pickmask(self._pickmask)
        if len(self.hidden_component) > 0:
            self.hide_component(self.hidden_component)
        self.set_gl_hl_use_array_idx(self.get_gl_hl_use_array_idx())

        for artist in self._artists:
            artist.do_stencil_test = False
            artist.figobj = self
            artist.figobj_hl = []
            artist.set_zorder(self.getp('zorder'))
            self._objs.append(weakref.ref(artist))
            if self.getvar('cz'):
                cax = self.get_caxisparam()
                if cax is None:
                    dprint1('Error: cax is None')
                cax.set_crangeparam_to_artist(artist)

    def switch_scale(self, level='fine'):
        return

    def del_artist(self, artist=None, delall=False):
        # if (len(self._artists) == 1 and
        #    self._artists[0] != self._fine_artist):
        #   self.switch_scale(level='fine')

        artistlist = self._artists
        self.store_loaded_property()

        if len(artistlist) != 0:
            self.highlight_artist(False, artistlist)
            container = self.get_container()
            for a in artistlist:
                try:
                    a.remove()
                except:
                    dprint1("remove failed")

        #self._fine_artist = None
        super(FigSolid, self).del_artist(artistlist)

    def get_mappable(self):
        if self.getvar('cz'):
            return [a for a in self._artists if isinstance(a, ScalarMappable)]
        else:
            return []

    def highlight_artist(self, val, artist=None):
        from ifigure.matplotlib_mod.art3d_gl import Poly3DCollectionGL
        figure = self.get_figpage()._artists[0]
        ax = self.get_figaxes()
        if artist is None:
            alist = self._artists
        else:
            alist = artist

        if val == True:
            if self._parent is None:
                return
            container = self.get_container()
            if container is None:
                return

            de = self.get_data_extent()
            x = (de[0], de[1], de[1], de[0], de[0])
            y = (de[2], de[2], de[3], de[3], de[2])

            facecolor = 'k'
            if isinstance(alist[0], Poly3DCollectionGL):
                hl = alist[0].add_hl_mask()
            else:
                hl = []

            for item in hl:
                alist[0].figobj_hl.append(item)

        else:
            for a in alist:
                if len(a.figobj_hl) == 0:
                    continue
                for hl in a.figobj_hl:
                    hl.remove()
                a.figobj_hl = []

    def set_cmap(self, value, a):
        a.set_cmap(cm.get_cmap(value))
        self.setp('cmap', value)
        ca = self.get_caxisparam()
        ca.set_cmap(value)
        if self.has_cbar():
            ca.update_cb()

    def get_cmap(self, a=None):
        ca = self.get_caxisparam()
        return ca.cmap

    def set_alpha(self, value, a):
        a.set_alpha(value)
        self.setp('alpha', value)
#        self.setp('cmap', value)
#        self.set_cmap(self.get_cmap(a), a)

    def get_alpha(self, a=None):
        return self.getp('alpha')

    def set_edgecolor(self, value, a):
        self.setp('edgecolor', value)
        a.set_edgecolor([value])

    def get_edgecolor(self, a=None):
        return self.getp('edgecolor')

    def set_facecolor(self, value, a):
        if self.getvar('cz'):
            return
        if value == 'disabled':
            return
        if isinstance(value, str):
            value = cc.to_rgba(value)
        alpha = self.getp('alpha')
        if alpha is None:
            alpha = 1.0
        value = (value[0], value[1], value[2], alpha)
        self.setp('facecolor', value)
        a.set_facecolor([value])

    def get_facecolor(self, a=None):
        if self.getvar('cz'):
            return 'disabled'
        return self.getp('facecolor')

    def set_linewidth(self, value, a=None):
        self.setp('linewidth', value)
        if len(self._artists) > 0:
            if a is None:
                a = self._artists[0]
            a.set_linewidth(value)

    def get_linewidth(self, a=None):
        return self.getp('linewidth')

    def set_shade(self, value, a):
        self.setvar('shade', value)
        self.reset_artist()

    def get_shade(self, a=None):
        return self.getvar('shade')

#
#   def hit_test
#
    def picker_a(self, artist, evt):
        axes = artist.axes
        if axes is None:
            return False, {}
        hit, extra = artist.contains(evt)

        if hit:
            return True, {}
        else:
            return False, {}

    def get_data_extent(self):
        if self._data_extent is not None:
            return self._data_extent
        x, y, z = self._eval_xyz()
        self._data_extent = [np.min(x), np.max(x),
                             np.min(y), np.max(y),
                             np.min(z), np.max(z)]
        return self._data_extent

    def get_xrange(self, xrange=[None, None], scale='linear'):
        x, y, z = self._eval_xyz()
        if x is None:
            return xrange
        if scale == 'log':
            x = mask_negative(x)
        return self._update_range(xrange, [np.nanmin(x), np.nanmax(x)])

    def get_yrange(self, yrange=[None, None], xrange=[None, None], scale='linear'):
        x, y, z = self._eval_xyz()
        if y is None:
            return yrange
        if scale == 'log':
            y = mask_negative(y)
        return self._update_range(yrange, (np.nanmin(y), np.nanmax(y)))

    def get_zrange(self, zrange=[None, None],
                   xrange=[None, None],
                   yrange=[None, None],
                   scale='linear'):
        x, y, z = self._eval_xyz()
        if z is None:
            return zrange
        zrange = self._update_range(zrange,
                                    (np.nanmin(z), np.nanmax(z)))

        return zrange

    def get_crange(self, crange=[None, None],
                   xrange=[None, None],
                   yrange=[None, None],
                   scale='linear'):
        cdata = self.getvar('cdata')
        cz = self.getvar('cz')
        if not cz:
            return crange
        if cdata is None:
            x, y, z = self._eval_xyz()
            crange = self._update_range(crange,
                                        (np.nanmin(z), np.nanmax(z)))

        else:
            if np.iscomplexobj(cdata):
                tmp = np.max(np.abs(cdata))
                crange = self._update_range(crange,
                                            (-tmp, tmp))

            else:
                crange = self._update_range(crange,
                                            (np.nanmin(cdata), np.nanmax(cdata)))

        return crange

    @classmethod
    def _saveload_names(self):
        return {'v', }

    def save_data2(self, data=None):
        def check(obj, name):
            if not isinstance(obj.getp(name), np.ndarray):
                return False
            if not isinstance(obj.getvar(name), np.ndarray):
                return False
            return obj.getp(name) is obj.getvar(name)

        if data is None:
            data = {}
        names = self._saveload_names()
        var = {name: check(self, name) for name in names}
        for name in names:
            if not var[name]:
                var[name+'data'] = self.getp(name)
        var['hidden'] = self.hidden_component

        data['FigSolid'] = (1, var)
        data = super(FigSolid, self).save_data2(data)
        return data

    def load_data2(self, data):
        d = data['FigSolid']
        super(FigSolid, self).load_data2(data)
        var = d[1]

        names = self._saveload_names()
        self.hide_component(var.pop('hidden', []))
        for name in names:
            if var[name]:
                self.setp(name,
                          self.getvar(name))
            else:
                self.setp(name, var[name+'data'])

    def _eval_xyz(self):
        v = self._eval_v()
        if v is None:
            return None, None, None
        return v[..., 0], v[..., 1], v[..., 2]

    def _eval_v(self):
        if self.getp('use_var'):
            success = self.handle_use_var()
            if not success:
                return None
        return self.getp("v")

    def get_export_val(self, a):
        val = {"vertices":  self.getp("v"), }
        if self.getvar('idxset') is not None:
            val['idxset'] = self.getvar("idxset")
        if self.getvar('cz') and self.getvar("cdata") is not None:
            val['cdata'] = self.getvar("cdata")
        return val

    def canvas_menu(self):
        m = FigObj.canvas_menu(self)
        m2 = GLCompound.canvas_menu(self)
        if self.get_gl_hl_use_array_idx():
            if len(self.getSelectedIndex()) > 0:
                m3 = [("Generate subset plot from selection",  self.onPlotSubsetS, None),
                      ("Generate subset plot",  self.onPlotSubsetV, None), ]
            else:
                m3 = [("Generate subset plot",  self.onPlotSubsetV, None), ]
            return m[:1]+m2+m3+m[1:]
        else:
            return m[:1]+m2+m[1:]

    def _on_plot_subset(self, cmp):
        '''
        create subset plot
        '''
        subset = self.get_subset(cmp)
        if len(subset) == 4:
            v, idx, cdata, s2 = subset
        else:
            v, idx, cdata = subset

        kwargs = {}
        arr = self.getvar('array_idx')
        kwargs['array_idx'] = arr[np.in1d(arr, cmp)]
        if cdata is None:
            kwargs['cz'] = False
        else:
            kwargs['cz'] = True
            kwargs['cdata'] = cdata

        from ifigure.interactive import figure
        win = figure()
        win.threed('on')
        win.solid(v, idx, **kwargs)

    def onPlotSubsetV(self, evt):
        '''
        create subset plot from visible part
        '''
        cmp = self.shown_component
        self._on_plot_subset(cmp)

    def onPlotSubsetS(self, evt):
        '''
        create subset plot
        '''
        cmp = self.getSelectedIndex()
        self._on_plot_subset(cmp)

#
# change rotation center in 3D plot
#

    def onSetRotCenter(self, evt):
        sel = self.getSelectedIndex()
        if len(sel) == 0:
            sel = self.shown_component

        if len(sel) > 0:
            vv = self.get_subset(sel)[0]

            cc = np.mean(vv, 0)
            axes = self._artists[0].axes
            axes._gl_rot_center = cc
            axes._gl_use_rot_center = True
            evt.Skip()
        else:
            super(FigSolid, self).onSetRotCenter(evt)



import wx
import numpy as np
from ifigure.utils.cbook import ProcessKeywords, LoadImageFile
from ifigure.widgets.undo_redo_history import GlobalHistory, UndoRedoFigobjMethod
from ifigure.utils.edit_list import DialogEditList
from ifigure.mto.fig_control import FigControl
from ifigure.mto.fig_axspan import FigAxspan
from ifigure.mto.fig_obj import set_mpl_all, get_mpl_first


class FigAxspanC(FigAxspan, FigControl):
    def __new__(cls, *args, **kargs):
        draggable, kargs = ProcessKeywords(kargs, 'draggable', True)
        dragmode, kargs = ProcessKeywords(kargs, 'dragmodee', 'independent')
#        dragmode, kargs = ProcessKeywords(kargs, 'alpha', 1.0)
#        dragmode, kargs = ProcessKeywords(kargs, 'alpha_disabled', 0.5)
        enabled_point, kargs = ProcessKeywords(kargs, 'enabled_point', 'all')
        obj = FigAxspan.__new__(cls, *args, **kargs)
        if obj is not None:
            obj.setvar('draggable', draggable)
            obj.setvar('dragmode', dragmode)
#            obj.setvar('alpha', alpha)
#            obj.setvar('alpha_disabled', alpha_disabled)
            if not obj.hasvar('enabled_point'):
                obj.setvar('enabled_point', enabled_point)
        return obj

    def __init__(self, *args, **kargs):
        FigAxspan.__init__(self, *args, **kargs)
        FigControl.__init__(self)
        self._figc_hit = -1
        self._enable_a = None

    @classmethod
    def get_namebase(self):
        return 'axspanc'

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = LoadImageFile(path, 'plotc.png')
        return [idx1]

    @classmethod
    def attr_in_file(self):
        return ['enabled_point']+super(FigAxspanC, self).attr_in_file()

    def picker_a(self, a, evt):
        hit, extra = FigAxspan.picker_a(self, a, evt)
        if hit:
            self._figc_hit = self._artists.index(extra['child_artist'])
            self._drag_backup = a.get_xy()
        else:
            self._figc_hit = -1
        return hit, extra

    def drag(self, a, evt):
        if self.getvar('dragmode') == 'independent':
            FigAxspan.drag(self, a, evt, idx=self._figc_hit)
        else:
            FigAxspan.drag(self, a, evt)

    def dragdone(self, a, evt):
        if self.getvar('dragmode') == 'independent':
            FigAxspan.dragdone(self, a, evt, idx=self._figc_hit)
        else:
            FigAxspan.dragdone(self, a, evt)

    def canvas_menu(self):
        m = [('+Add...', None, None),
             ('X span', self.onAddXSpan, None),
             ('Y span', self.onAddYSpan, None),
             ('!', None, None), ]
        if self._figc_hit != -1:
            if len(self.getp('x'))+len(self.getp('y')) > 1:
                m = m + [('Delete Span', self.onDelSpan, None), ]
            if self.getp('enabled_point')[self._figc_hit]:
                m = m + [('Edit Data...', self.onEditSpan, None),
                         #('Disalbe Span...', self.onEnableSpan, None),
                         ]

            else:
                m = m + [
                    ('Edit Data...', self.onEditSpan, None),
                    #                        ('Enable Span', self.onEnablePoint, None),
                ]
        return m + FigAxspan.canvas_menu(self)

    def onEditSpan(self, evt):
        x, y = self._eval_xy()
        if self._figc_hit < x.size:
            l = [['x1', str(x[self._figc_hit, 0]), 0, None],
                 ['x2', str(x[self._figc_hit, 1]), 0, None], ]
        else:
            l = [['y1', str(y[self._figc_hit-x.size, 0]), 0, None],
                 ['y2', str(y[self._figc_hit-x.size, 1]), 0, None], ]
        window = evt.GetEventObject().GetTopLevelParent()
        value = DialogEditList(l, modal=True,
                               style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER,
                               tip=None,
                               parent=window)
        if value[0]:
            v = value[1]
        else:
            return
        x = self.getp('x').copy()
        y = self.getp('y').copy()
        if self._figc_hit < x.size:
            x[self._figc_hit, 0] = float(v[0])
            x[self._figc_hit, 1] = float(v[1])
        else:
            y[self._figc_hit - x.size, 0] = v[0]
            y[self._figc_hit - x.size, 1] = v[1]
        action = UndoRedoFigobjMethod(self._artists[0],
                                      'data',
                                      (x, y))
        window = evt.GetEventObject().GetTopLevelParent()
        GlobalHistory().get_history(window).make_entry([action])
        return 1

    def onEnableSpan(self, evt):
        pass

    def onDisableSpan(self, evt):
        pass

    def onDelSpan(self, evt):
        x, y = self._eval_xy()
        x = x.copy()
        y = y.copy()
        if self._figc_hit < x.size:
            x = np.array([xx for i, xx in enumerate(x) if i != self._figc_hit])
        else:
            y = np.array([xx for i, xx in enumerate(
                y) if i != self._figc_hit-len(x)])
        action = UndoRedoFigobjMethod(self._artists[0],
                                      'data',
                                      (x, y))
        window = evt.GetEventObject().GetTopLevelParent()
        GlobalHistory().get_history(window).make_entry([action])
        return 1

    def onAddXSpan(self, evt):
        l = [['x1', 0, 400, None],
             ['x2', 1, 400, None], ]
        window = evt.GetEventObject().GetTopLevelParent()
        value = DialogEditList(l, modal=True,
                               style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER,
                               tip=None,
                               parent=window)
        if value[0]:
            v = value[1]
        else:
            return
        x = self.getp('x').copy()
        y = self.getp('y').copy()
        x = np.array([xx for xx in x] + [v])
        action = UndoRedoFigobjMethod(self._artists[0],
                                      'data',
                                      (x, y))
        window = evt.GetEventObject().GetTopLevelParent()
        GlobalHistory().get_history(window).make_entry(
            [action], menu_name='add span')
        return 1

    def onAddYSpan(self, evt):
        l = [['y1', 0, 400, None],
             ['y2', 1, 400, None], ]
        window = evt.GetEventObject().GetTopLevelParent()
        value = DialogEditList(l, modal=True,
                               style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER,
                               tip=None,
                               parent=window)
        if value[0]:
            v = value[1]
        else:
            return
        x = self.getp('x').copy()
        y = self.getp('y').copy()
        y = np.array([xx for xx in y] + [v])
        action = UndoRedoFigobjMethod(self._artists[0],
                                      'data',
                                      (x, y))
        window = evt.GetEventObject().GetTopLevelParent()
        GlobalHistory().get_history(window).make_entry(
            [action], menu_name='add span')
        return 1

    def set_alpha(self, value, a):
        set_mpl_all(self._artists, 'alpha', value)
#            if self.getp('enabled_point')[self._figc_hit]:

    def get_alpha(self, a):
        return get_mpl_first(self._artists, 'alpha')

    def set_enabled_point(self, value, a):
        self.setp('enabled_point', value)
        self.set_update_artist_request()
        self.get_figaxes().set_bmp_update(False)

    def get_enabled_point(self, a):
        return self.getp('enabled_point')

    def set_data(self, value, a):
        FigAxspan.set_data(self, value, a)
        self._data_extent = None
        self.set_update_artist_request()
        self.call_control_changed_callback()

    def get_data(self, a=None):
        return (self.getp('x').copy(), self.getp('y').copy(), self.getp('enabled_point')[:])

    def save_data2(self, data=None):
        var = {'enabled_point': self.getp('enabled_point')}
        data['FigAxspanC'] = (1, var)
        data = super(FigAxspanC, self).save_data2(data)
        return data

    def load_data2(self, data):
        super(FigAxspanC, self).load_data2(data)
        if not 'FigAxspanC' in data:
            self.setp('enabled_point',
                      [True]*(self.getp('x').size + self.getp('y').size))
        else:
            d = data['FigAxspanC'][1]
            self.setp('enabled_point', d['enabled_point'])



import numpy as np
from ifigure.interactive import figure

'''
   This exapmle shows how to write a function in Script object.
   Script object can be called as if it is a function
  
       proj.xxx.yyy.zzz.obj(x, y, keyword = 'xxxx')

   Note following keywords are reserved by piScope
 
   model
   ans
   obj
   top
   app
   wdir
   param
   ans
   args
   kwargs
'''


def func(x, y, keyword=True):
    pass


ans(func(*args, **kwargs))



"""
   RendererGL

"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import six

import threading
import numpy as np

from matplotlib import verbose, rcParams
from matplotlib.backend_bases import RendererBase,\
    FigureManagerBase, FigureCanvasBase
from matplotlib.cbook import is_string_like, maxdict, restrict_dict
from matplotlib.figure import Figure
from matplotlib.font_manager import findfont
from matplotlib.ft2font import FT2Font, LOAD_FORCE_AUTOHINT, LOAD_NO_HINTING, \
    LOAD_DEFAULT, LOAD_NO_AUTOHINT
from matplotlib.mathtext import MathTextParser
from matplotlib.path import Path
from matplotlib.transforms import Bbox, BboxBase

#from matplotlib.backends._backend_agg import RendererAgg as _RendererAgg
from matplotlib import _png

try:
    from PIL import Image
    _has_pil = True
except ImportError:
    _has_pil = False

backend_version = 'v2.2'


def get_hinting_flag():
    mapping = {
        True: LOAD_FORCE_AUTOHINT,
        False: LOAD_NO_HINTING,
        'either': LOAD_DEFAULT,
        'native': LOAD_NO_AUTOHINT,
        'auto': LOAD_FORCE_AUTOHINT,
        'none': LOAD_NO_HINTING
    }
    return mapping[rcParams['text.hinting']]


class RendererGLCore(object):
    def __init__(self, width, height, dpi, debug=False):
        self.dpi = dpi
        self.width = width
        self.height = height
        self.debug = debug
        self.buf = np.zeros((width, height, 4), dtype=np.uint8)

    def __getattr__(self, attr):
        print("RendercoreGL "+attr + " is called")
        return self.method

    def method(self, *args, **kwargs):
        pass

    def buffer_rgba(self):
        return self.buf.tobytes()


class RendererGL(RendererBase):
    """
    The renderer handles all the drawing primitives using a graphics
    context instance that controls the colors/styles
    """
    debug = 1

    # we want to cache the fonts at the class level so that when
    # multiple figures are created we can reuse them.  This helps with
    # a bug on windows where the creation of too many figures leads to
    # too many open file handles.  However, storing them at the class
    # level is not thread safe.  The solution here is to let the
    # FigureCanvas acquire a lock on the fontd at the start of the
    # draw, and release it when it is done.  This allows multiple
    # renderers to share the cached fonts, but only one figure can
    # draw at at time and so the font cache is used by only one
    # renderer at a time

    lock = threading.RLock()
    _fontd = maxdict(50)

    def __init__(self, width, height, dpi):
        if __debug__:
            verbose.report('RendererAgg.__init__', 'debug-annoying')
        RendererBase.__init__(self)
        self.texd = maxdict(50)  # a cache of tex image rasters

        self.dpi = dpi
        self.width = width
        self.height = height
        if __debug__:
            verbose.report('RendererAgg.__init__ width=%s, height=%s' %
                           (width, height), 'debug-annoying')
        self._renderer = RendererGLCore(
            int(width), int(height), dpi, debug=False)
        self._filter_renderers = []

        if __debug__:
            verbose.report('RendererAgg.__init__ _RendererAgg done',
                           'debug-annoying')

        self._update_methods()
        self.mathtext_parser = MathTextParser('Agg')

        self.bbox = Bbox.from_bounds(0, 0, self.width, self.height)
        if __debug__:
            verbose.report('RendererGL.__init__ done',
                           'debug-annoying')

    def __getstate__(self):
        # We only want to preserve the init keywords of the Renderer.
        # Anything else can be re-created.
        return {'width': self.width, 'height': self.height, 'dpi': self.dpi}

    def __setstate__(self, state):
        self.__init__(state['width'], state['height'], state['dpi'])

    def _get_hinting_flag(self):
        if rcParams['text.hinting']:
            return LOAD_FORCE_AUTOHINT
        else:
            return LOAD_NO_HINTING

    # for filtering to work with rasterization, methods needs to be wrapped.
    # maybe there is better way to do it.
    def draw_markers(self, *kl, **kw):
        return self._renderer.draw_markers(*kl, **kw)

    def draw_path_collection(self, *kl, **kw):
        return self._renderer.draw_path_collection(*kl, **kw)

    def _update_methods(self):
        # self.draw_path = self._renderer.draw_path  # see below
        #self.draw_markers = self._renderer.draw_markers
        #self.draw_path_collection = self._renderer.draw_path_collection
        self.draw_quad_mesh = self._renderer.draw_quad_mesh
        self.draw_gouraud_triangle = self._renderer.draw_gouraud_triangle
        self.draw_gouraud_triangles = self._renderer.draw_gouraud_triangles
        self.draw_image = self._renderer.draw_image
        self.copy_from_bbox = self._renderer.copy_from_bbox
        self.tostring_rgba_minimized = self._renderer.tostring_rgba_minimized

    def draw_path(self, gc, path, transform, rgbFace=None):
        """
        Draw the path
        """
        nmax = rcParams['agg.path.chunksize']  # here at least for testing
        npts = path.vertices.shape[0]
        if (nmax > 100 and npts > nmax and path.should_simplify and
                rgbFace is None and gc.get_hatch() is None):
            nch = np.ceil(npts/float(nmax))
            chsize = int(np.ceil(npts/nch))
            i0 = np.arange(0, npts, chsize)
            i1 = np.zeros_like(i0)
            i1[:-1] = i0[1:] - 1
            i1[-1] = npts
            for ii0, ii1 in zip(i0, i1):
                v = path.vertices[ii0:ii1, :]
                c = path.codes
                if c is not None:
                    c = c[ii0:ii1]
                    c[0] = Path.MOVETO  # move to end of last chunk
                p = Path(v, c)
                self._renderer.draw_path(gc, p, transform, rgbFace)
        else:
            self._renderer.draw_path(gc, path, transform, rgbFace)

    def draw_mathtext(self, gc, x, y, s, prop, angle):
        """
        Draw the math text using matplotlib.mathtext
        """
        if __debug__:
            verbose.report('RendererGL.draw_mathtext',
                           'debug-annoying')
        ox, oy, width, height, descent, font_image, used_characters = \
            self.mathtext_parser.parse(s, self.dpi, prop)

        xd = descent * np.sin(np.deg2rad(angle))
        yd = descent * np.cos(np.deg2rad(angle))
        x = np.round(x + ox + xd)
        y = np.round(y - oy + yd)
        self._renderer.draw_text_image(font_image, x, y + 1, angle, gc)

    def draw_text(self, gc, x, y, s, prop, angle, ismath=False, mtext=None):
        """
        Render the text
        """
        if __debug__:
            verbose.report('RendererGL.draw_text', 'debug-annoying')

        if ismath:
            return self.draw_mathtext(gc, x, y, s, prop, angle)

        flags = get_hinting_flag()
        font = self._get_agg_font(prop)
        if font is None:
            return None
        if len(s) == 1 and ord(s) > 127:
            font.load_char(ord(s), flags=flags)
        else:
            # We pass '0' for angle here, since it will be rotated (in raster
            # space) in the following call to draw_text_image).
            font.set_text(s, 0, flags=flags)
        font.draw_glyphs_to_bitmap(antialiased=rcParams['text.antialiased'])
        d = font.get_descent() / 64.0
        # The descent needs to be adjusted for the angle
        xd = -d * np.sin(np.deg2rad(angle))
        yd = d * np.cos(np.deg2rad(angle))

        # print x, y, int(x), int(y), s
        self._renderer.draw_text_image(
            font.get_image(), np.round(x - xd), np.round(y + yd) + 1, angle, gc)

    def get_text_width_height_descent(self, s, prop, ismath):
        """
        get the width and height in display coords of the string s
        with FontPropertry prop

        # passing rgb is a little hack to make caching in the
        # texmanager more efficient.  It is not meant to be used
        # outside the backend
        """
#        if rcParams['text.usetex']:
        if 1:
            # todo: handle props
            size = prop.get_size_in_points()
            texmanager = self.get_texmanager()
            fontsize = prop.get_size_in_points()
            w, h, d = texmanager.get_text_width_height_descent(s, fontsize,
                                                               renderer=self)
            return w, h, d

        if ismath:
            ox, oy, width, height, descent, fonts, used_characters = \
                self.mathtext_parser.parse(s, self.dpi, prop)
            return width, height, descent

        flags = get_hinting_flag()
        font = self._get_agg_font(prop)
        # the width and height of unrotated string
        font.set_text(s, 0.0, flags=flags)
        w, h = font.get_width_height()
        d = font.get_descent()
        w /= 64.0  # convert from subpixels
        h /= 64.0
        d /= 64.0
        return w, h, d

    def draw_tex(self, gc, x, y, s, prop, angle, ismath='TeX!', mtext=None):
        # todo, handle props, angle, origins
        size = prop.get_size_in_points()

        texmanager = self.get_texmanager()
        key = s, size, self.dpi, angle, texmanager.get_font_config()
        im = self.texd.get(key)
        if im is None:
            Z = texmanager.get_grey(s, size, self.dpi)
            Z = np.array(Z * 255.0, np.uint8)

        w, h, d = self.get_text_width_height_descent(s, prop, ismath)
        xd = d * np.sin(np.deg2rad(angle))
        yd = d * np.cos(np.deg2rad(angle))
        x = np.round(x + xd)
        y = np.round(y + yd)

        self._renderer.draw_text_image(Z, x, y, angle, gc)

    def get_canvas_width_height(self):
        'return the canvas width and height in display coords'
        return self.width, self.height

    def _get_agg_font(self, prop):
        """
        Get the font for text instance t, cacheing for efficiency
        """
        if __debug__:
            verbose.report('RendererGL._get_agg_font',
                           'debug-annoying')
        return None

        key = hash(prop)
        font = RendererAgg._fontd.get(key)

        if font is None:
            fname = findfont(prop)
            font = RendererAgg._fontd.get(fname)
            if font is None:
                font = FT2Font(
                    fname,
                    hinting_factor=rcParams['text.hinting_factor'])
                RendererAgg._fontd[fname] = font
            RendererAgg._fontd[key] = font

        font.clear()
        size = prop.get_size_in_points()
        font.set_size(size, self.dpi)

        return font

    def points_to_pixels(self, points):
        """
        convert point measures to pixes using dpi and the pixels per
        inch of the display
        """
        if __debug__:
            verbose.report('RendererGL.points_to_pixels',
                           'debug-annoying')
        return points*self.dpi/72.0

    def tostring_rgb(self):
        if __debug__:
            verbose.report('RendererGL.tostring_rgb',
                           'debug-annoying')
        return self._renderer.tostring_rgb()

    def tostring_argb(self):
        if __debug__:
            verbose.report('RendererGL.tostring_argb',
                           'debug-annoying')
        return self._renderer.tostring_argb()

    def buffer_rgba(self):
        if __debug__:
            verbose.report('RendererGL.buffer_rgba',
                           'debug-annoying')
        return self._renderer.buffer_rgba()

    def clear(self):
        self._renderer.clear()

    def option_image_nocomposite(self):
        # It is generally faster to composite each image directly to
        # the Figure, and there's no file size benefit to compositing
        # with the Agg backend
        return True

    def option_scale_image(self):
        """
        agg backend support arbitrary scaling of image.
        """
        return True

    def restore_region(self, region, bbox=None, xy=None):
        return RendererBase.restore_region(self, region, bbox=None, xy=None)
        """
        Restore the saved region. If bbox (instance of BboxBase, or
        its extents) is given, only the region specified by the bbox
        will be restored. *xy* (a tuple of two floasts) optionally
        specifies the new position (the LLC of the original region,
        not the LLC of the bbox) where the region will be restored.

        >>> region = renderer.copy_from_bbox()
        >>> x1, y1, x2, y2 = region.get_extents()
        >>> renderer.restore_region(region, bbox=(x1+dx, y1, x2, y2),
        ...                         xy=(x1-dx, y1))

        """
        if bbox is not None or xy is not None:
            if bbox is None:
                x1, y1, x2, y2 = region.get_extents()
            elif isinstance(bbox, BboxBase):
                x1, y1, x2, y2 = bbox.extents
            else:
                x1, y1, x2, y2 = bbox

            if xy is None:
                ox, oy = x1, y1
            else:
                ox, oy = xy

            self._renderer.restore_region2(region, x1, y1, x2, y2, ox, oy)

        else:
            self._renderer.restore_region(region)

    def start_filter(self):
        return RendererBase.start_filter(self)
        """
        Start filtering. It simply create a new canvas (the old one is saved).
        """
        self._filter_renderers.append(self._renderer)
        self._renderer = _RendererAgg(int(self.width), int(self.height),
                                      self.dpi)
        self._update_methods()

    def stop_filter(self, post_processing):
        return RendererBase.stop_filter(self, post_processing)
        """
        Save the plot in the current canvas as a image and apply
        the *post_processing* function.

           def post_processing(image, dpi):
             # ny, nx, depth = image.shape
             # image (numpy array) has RGBA channels and has a depth of 4.
             ...
             # create a new_image (numpy array of 4 channels, size can be
             # different). The resulting image may have offsets from
             # lower-left corner of the original image
             return new_image, offset_x, offset_y

        The saved renderer is restored and the returned image from
        post_processing is plotted (using draw_image) on it.
        """

        # WARNING.
        # For agg_filter to work, the rendere's method need
        # to overridden in the class. See draw_markers, and draw_path_collections

        from matplotlib._image import fromarray

        width, height = int(self.width), int(self.height)

        buffer, bounds = self._renderer.tostring_rgba_minimized()

        l, b, w, h = bounds

        self._renderer = self._filter_renderers.pop()
        self._update_methods()

        if w > 0 and h > 0:
            img = np.fromstring(buffer, np.uint8)
            img, ox, oy = post_processing(img.reshape((h, w, 4)) / 255.,
                                          self.dpi)
            image = fromarray(np.flipud(img), 1)
#            image.flipud_out()

            gc = self.new_gc()
            self._renderer.draw_image(gc,
                                      l+ox, height - b - h + oy,
                                      image)



from __future__ import print_function
import struct


def idltype2unitsize(x):
    # x is type code in IDL
    # returns unitsize
    units = {"1": 1,
             "2": 2,
             "3": 4,
             "4": 4,
             "5": 8,
             "6": 8,
             "7": 1,
             "9": 16,
             "12": 2,
             "13": 4,
             "14": 8,
             "15": 8,
             }
    return units[str(x)]


def idltype2code(x):
    # x is type code in IDL
    # returns unitsize
    units = {'1': "b",
             '2': "h",
             '3': "l",
             '4': "f",
             '5': "d",
             '6': "f",
             '7': "s",
             '9': "d",
             '12': "H",
             '13': "L",
             '14': "q",
             '15': "Q",
             }
    return units[str(x)]


def idl2python_type(s):

    #   IDL definition of type code
    #   0:  undefined
    #   1:  byte
    #   2:  integer
    #   3:  long
    #   4:  float
    #   5:  double
    #   6:  complex
    #   7:  string
    #   8:  structure
    #   9:  double complex
    #  10:  pointer
    #  11:  obj ref
    #  12:  unsigned int
    #  13:  unsigned long
    #  14:  long64
    #  15:  unsigned long64

    #  size(a) in IDL returns
    #    [ndim, length of each dimension, type-code, total number of element]

    idltype = s[len(s)-2]

    unit = 0
    code = ""
    iscomplex = 0

    unitsize = idltype2unitsize(idltype)
    code = idltype2code(idltype)
    if idltype == 6:
        iscomplex = 1
    if idltype == 9:
        iscomplex = 1

    return [unitsize, code, iscomplex]


def read_structure(fin):

    val = {}
    x = struct.unpack('l', fin.read(4))
    n = x[0]
    if n == 0:
        return val

    k = 1

    while k <= n:
        btag = struct.unpack('l', fin.read(4))

        tag = struct.unpack(str(btag[0])+"s", fin.read(btag[0]))
        print(tag)
        ndim = struct.unpack('l', fin.read(4))

        sz = struct.unpack(str(ndim[0]+2)+'l', fin.read(4*(ndim[0]+2)))

        nele = sz[len(sz)-1]
        print(nele)
        [unit, code, iscomplex] = idl2python_type(sz)
        if iscomplex:
            array = struct.unpack(str(nele*2)+code,  fin.read(unit*2*nele))
        else:
            array = struct.unpack(str(nele)+code,  fin.read(unit*nele))

        print(array)
        # should modify the dim and complex here
        if nele == 1:
            val[tag[0].upper()] = array[0]
        else:
            val[tag[0].upper()] = array[0:]
        k = k+1

    return val



from __future__ import print_function
import os
import numpy as np
import time
import wx
import weakref

import ifigure.utils.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('canvas_common')

#
#  OpenGL extention
#
try:
    from wx import glcanvas
except ImportError:
    pass
try:
    # The Python OpenGL package can be found at
    # http://PyOpenGL.sourceforge.net/
    from OpenGL.GL import *
    #from OpenGL.GLUT import *
    from OpenGL.GLU import *
    from OpenGL.GL import shaders
    from OpenGL.arrays import vbo

    class myvbo(vbo.VBO):
        pass
    haveOpenGL = True
except ImportError:
    haveOpenGL = False

    
near_clipping = 45.    # must be float (default 8) = A
camera_distance = 50.  # must be float (default 10) = B
# I haven't check relations between these. Apparently, (B-A)/B = 1/10.?

view_scale = 1.

multisample = 1
multisample_init_done = False

basedir = os.path.dirname(__file__)


def compile_file(file, mode):
    fid = open(os.path.join(basedir, file), 'r')
    prog = ''.join(fid.readlines())
    fid.close()
    pl = shaders.compileShader(prog, mode)
    return pl


def get_vbo(data, *args, **kwargs):
    vbo = myvbo(data, *args, **kwargs)
    vbo.need_update = False
    return vbo


def read_glmatrix(mode):
    #    a = (GLfloat * 16)()
    return np.transpose(glGetFloatv(mode))
#    return np.transpuse(np.array(list(a)).reshape(4,4))


def define_unform(shader, name):
    shader.uniform_loc[name] = glGetUniformLocation(shader, name)


def define_attrib(shader, name):
    shader.attrib_loc[name] = glGetAttribLocation(shader, name)


def check_framebuffer(message, mode=GL_FRAMEBUFFER):
    # list up possible errors
    attrs = [GL_FRAMEBUFFER_COMPLETE, GL_FRAMEBUFFER_UNSUPPORTED, GL_INVALID_ENUM]
    for x in dir(OpenGL.GL):
        if x.startswith('GL_FRAMEBUFFER_INCOMPLETE'):
            attrs.append(getattr(OpenGL.GL,x))
    #for a in attrs:
    #    print(a, int(a))

    check = glCheckFramebufferStatus(mode)
    if int(check) != int(attrs[0]):
        print('Framebuffer imcomplete (' + message + ')')
        for x in attrs:
            if int(check) == int(x):
                print(x)
        return False

    # print "test sample", glGetIntegerv(GL_SAMPLE_BUFFERS)
    return True

def frustum(left, right, bottom, top, zNear, zFar, view_scale=1):
    dx = right - left
    dy = top - bottom
    A = (right + left) / (right - left)
    B = (top + bottom) / (top - bottom)
    C = -(zFar + zNear) / (zFar - zNear)
    D = - (2*zFar * zNear) / (zFar - zNear)
    M = np.array([[2*zNear/dx*view_scale, 0,           A, 0],
                  [0,          2*zNear/dy*view_scale,  B, 0],
                  [0,          0,           C, D],
                  [0,          0,          -1, 0]])
    return M


def ortho(left, right, bottom, top, zNear, zFar, view_scale=1):
    dx = right - left
    dy = top - bottom
    dz = zFar - zNear
    tx = - (right + left) / (right - left)
    ty = - (top + bottom) / (top - bottom)
    tz = - (zFar + zNear) / (zFar - zNear)
    return np.array([[2/dx*view_scale, 0,     0,     tx],
                     [0,    2/dy*view_scale,  0,     ty],
                     [0,    0,     -2/dz,    tz],
                     [0,    0,     0,        1.]])


def wait_gl_finish(method):
    @wraps(method)
    def method2(self, *args, **kargs):
        method(self, *args, **kargs)
        glFinish()
    return method


def check_gl_error():
    error = glGetError()
    if error != 0:
        print(("GL error ", error))


class vbos_dict(dict):
    def __del__(self, *args, **kwargs):
        if 'im' in self:
            if self['im'] is not None:
                dprint2('deleteing texture', self['im'])
                glDeleteTextures(self['im'])
                self['im'] = None
        return



from __future__ import print_function
import numpy as np


def image_demo():
    from ifigure.interactive import image, threed, figure, hold

    figure(gl=True)
    threed('on')
    hold('on')
    X = np.linspace(-5, 5, 40)
    Y = np.linspace(-5, 5, 40)
    X, Y = np.meshgrid(X, Y)
    R = np.sqrt(X**2 + Y**2)
    Z = np.sin(R)
    image(Z, cmap='coolwarm', im_center=[0, 0, 1])
    image(Z, cmap='coolwarm', im_center=[0, 0, 1],
          im_axes=([0, 0, 1], [0, 1, 0]))
    image(Z, cmap='coolwarm', im_center=[0, 0, 1],
          im_axes=([1/np.sqrt(2), 0, 1/np.sqrt(2)],
                   [0, 1, 0]))


def surf_demo(**kwargs):
    from ifigure.interactive import surf, threed, figure, hold, lighting

    figure()
    threed('on')
    scale = kwargs.pop("scale", 1.0)
#   X = np.arange(-5, 5, 0.25)
#   Y = np.arange(-5, 5, 0.25)
    X = np.linspace(-5, 5, 40)
    Y = np.linspace(-5, 5, 40)
    X, Y = np.meshgrid(X, Y)
    R = np.sqrt(X**2 + Y**2)
    Z = np.sin(R)*scale

    lighting(light=0.5, ambient=0.7)
    surf(X, Y, Z, cmap='coolwarm', shade='linear', **kwargs)
#  Debug to draw flat plane....
#   hold(1)
#   surf(np.array([-5.0,-5, -5.0]), np.array([-5, 0, 5]), np.array([[1,0, -1],[1, 0, -1],[1, 0, -1]]), cmap='coolwarm', **kwargs)
#   surf(np.array([5.0, 5, 5.0]), np.array([-5, 0, 5]), np.array([[1,0, -1],[1, 0, -1],[1, 0, -1]]), cmap='coolwarm', **kwargs)


def surf_demo2(**kwargs):
    from mpl_toolkits.mplot3d import Axes3D
    import numpy as np

    from ifigure.interactive import surf, threed, figure, isec, nsec

    figure(gl=True)
    nsec(3)
    isec(0)
    threed('on')
    u = np.linspace(0, 2 * np.pi, 100)
    v = np.linspace(0, np.pi, 100)

    x = 10 * np.outer(np.cos(u), np.sin(v))
    y = 10 * np.outer(np.sin(u), np.sin(v))
    z = 10 * np.outer(np.ones(np.size(u)), np.cos(v))
    surf(x, y, z, cz=False, facecolor='b', edgecolor='k')

    isec(1)
    threed('on')
    surf(x, y, z, cstride=4, cz=True,  edgecolor='k')

    isec(2)
    threed('on')
    surf(x, y, z,  cz=True, cdata=y, edgecolor=None)


def contour_demo(**kwargs):
    import mpl_toolkits.mplot3d.axes3d as axes3d
    from ifigure.interactive import contour, threed, figure
    X, Y, Z = axes3d.get_test_data(0.05)
    v = figure(gl=True)
    threed('on')
    contour(X, Y, Z)


def contourf_demo(**kwargs):
    import mpl_toolkits.mplot3d.axes3d as axes3d
    from ifigure.interactive import contourf, threed, figure
    X, Y, Z = axes3d.get_test_data(0.05)
    v = figure(gl=True)
    threed('on')
    contourf(X, Y, Z)

def contourf_demo2(**kwargs):
    from ifigure.interactive import contourf, threed, figure

    v = figure()
    v.threed('on')
    X = np.linspace(-5, 5, 300)
    Y = np.linspace(-5, 5, 300)
    X, Y = np.meshgrid(X, Y)
    R = np.sqrt(X**2 + Y**2)
    Z = np.cos(R*2)*np.exp(-R**2/10)
    #v.contourf(X, Y, Z, cmap='coolwarm')
    Z = np.hstack((Z,Z))
    v.contourf(Z, cmap='coolwarm')


def contour_demo2(**kwargs):
    import mpl_toolkits.mplot3d.axes3d as axes3d
    from ifigure.interactive import contourf, contour, threed, figure, hold
    X, Y, Z = axes3d.get_test_data(0.05)
    v = figure(gl=True)
    threed('on')
    hold('on')
    contour(X, Y, Z)
    contour(X, Y, Z, zdir='x', offset=-40)
    contour(X, Y, Z, zdir='y', offset= 40)
    contour(X, Y, Z, zdir='z', offset=-100)


def revolve_demo(**kwargs):
    m = np.linspace(0, np.pi*2, 30)
    r = 1.0 + 0.2*np.cos(m)
    z = 0.25 * np.sin(m)
    from ifigure.interactive import revolve, figure, threed
    v = figure(gl=True)
    threed('on')
    revolve(r, z, raxis=[0.5, 1], rtheta=[0, 2*np.pi/3])
    v.xlim(-1.5, 1.5)
    v.ylim(-1.5, 1.5)
    v.zlim(-1.5, 1.5)


def solid_stl_demo(**kwargs):
    from stl import mesh
    from ifigure.interactive import solid, figure, threed
    import ifigure
    import os
    from os.path import dirname
    path = dirname(dirname(dirname(ifigure.__file__)))
    mymesh = mesh.Mesh.from_file(
        os.path.join(path, 'example', 'D_antenna.stl'))
    v = figure()
    threed('on')
    solid(mymesh.vectors, alpha=0.5, **kwargs)


def solid_demo(**kwargs):
    '''
      solid_demo2(cz = True, linewidths = 1.0, edgecolor='red')
      solid_demo2(facecolor='b', linewidths = 1.0, edgecolor='red')

    '''
    from ifigure.interactive import solid, figure, threed, isec, nsec, lighting
    import ifigure
    import os

    # preparing the data, the same as mplot3d demo
    theta = np.linspace(0, 2.0 * np.pi, endpoint=True, num=50)
    r = np.linspace(0.01, 4.0, endpoint=True, num=50)
    # This is the Mobius mapping, taking a u, v pair and returning an x, y, z
    # triple
    THETA, R = np.meshgrid(theta, r)
    X = R*np.cos(THETA)
    Y = R*np.sin(THETA)
    Z = np.exp(- R*R)

    import matplotlib.tri as mtri
    tri = mtri.Triangulation(X.flatten(), Y.flatten())
    # Triangulate parameter space to determine the triangles

    if 'cz' in kwargs and kwargs['cz']:
        v = np.dstack((tri.x[tri.triangles],
                       tri.y[tri.triangles],
                       Z.flatten()[tri.triangles],))
        v = np.vstack((tri.x, tri.y, Z.flatten()))
        kwargs['cdata'] = 5*Z.flatten()[tri.triangles]
        kwargs['cdata'] = 5*Z.flatten()
        kwargs['cz'] = True

        # or #
        v = np.dstack((tri.x[tri.triangles],
                       tri.y[tri.triangles],
                       Z.flatten()[tri.triangles],
                       5 * Z.flatten()[tri.triangles],))
        v = np.vstack((tri.x, tri.y, Z.flatten(), 5*Z.flatten()))
    elif 'twod' in kwargs and kwargs['twod']:
        v = np.dstack((tri.x[tri.triangles],
                       tri.y[tri.triangles]),)
        v = np.vstack((tri.x, tri.y, ))
        kwargs.pop('twod')
        kwargs['zvalue'] = 1.0
    else:
        v = np.dstack((tri.x[tri.triangles],
                       tri.y[tri.triangles],
                       Z.flatten()[tri.triangles],))
        v = np.vstack((tri.x, tri.y, Z.flatten()))

    idxset = tri.triangles
    v = v.transpose()

    #print(v.shape, idxset.shape)
    viewer = figure()
    nsec(3)
    isec(0)
    threed('on')
    solid(v, idxset, edgecolor='k', facecolor='b')
    lighting(light=0.5, ambient=0.5)
    isec(1)
    threed('on')
    solid(v, idxset, cz=True)
    lighting(light=0.5, ambient=0.5)
    isec(2)
    threed('on')
    solid(v, idxset, cz=True, cdata=v[..., 0], shade='linear')
    lighting(light=0.5, ambient=0.5)


def triplots_demo(**kwargs):
    from matplotlib import cm
    import numpy as np
    from ifigure.interactive import figure, threed, trisurf, nsec, isec, lighting, tricontourf

    # preparing the data, the same as mplot3d demo
    n_angles = 36
    n_radii = 8
    radii = np.linspace(0.125, 1.0, n_radii)
    angles = np.linspace(0, 2*np.pi, n_angles, endpoint=False)
    angles = np.repeat(angles[..., np.newaxis], n_radii, axis=1)
    x = np.append(0, (radii*np.cos(angles)).flatten())
    y = np.append(0, (radii*np.sin(angles)).flatten())
    z = np.sin(-x*y)

    v = figure()
    nsec(4)
    isec(0)
    threed('on')
    trisurf(x, y, z, cmap=cm.jet, linewidth=0.2, cz=True)
    isec(1)
    threed('on')
    trisurf(x, y, z, linewidth=0.2,  color='b')
    lighting(light=0.5, ambient=0.5)
    isec(2)
    threed('on')
    trisurf(x, y, z*0, cmap=cm.jet, linewidth=0.2,
            cz=True, cdata=z, edgecolor=None)
    isec(3)
    threed('on')
    tricontourf(x, y, z)


def trisurf3d_demo2(**kwargs):
    import numpy as np
    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D
    import matplotlib.tri as mtri
    from ifigure.interactive import figure

    # preparing the data, the same as mplot3d demo
    u = (np.linspace(0, 2.0 * np.pi, endpoint=True, num=50)
         * np.ones((10, 1))).flatten()
    v = np.repeat(np.linspace(-0.5, 0.5, endpoint=True, num=10),
                  repeats=50).flatten()
    # This is the Mobius mapping, taking a u, v pair and returning an x, y, z
    # triple
    x = (1 + 0.5 * v * np.cos(u / 2.0)) * np.cos(u)
    y = (1 + 0.5 * v * np.cos(u / 2.0)) * np.sin(u)
    z = 0.5 * v * np.sin(u / 2.0)
    # Triangulate parameter space to determine the triangles
    tri = mtri.Triangulation(u, v)

    v = figure(gl=True)
    v.nsec(2)
    v.isec(0)
    v.threed('on')
    v.trisurf(x, y, z, triangles=tri.triangles, cmap=plt.cm.Spectral, cz=True)

    # First create the x and y coordinates of the points.
    n_angles = 36
    n_radii = 8
    min_radius = 0.25
    radii = np.linspace(min_radius, 0.95, n_radii)

    angles = np.linspace(0, 2*np.pi, n_angles, endpoint=False)
    angles = np.repeat(angles[..., np.newaxis], n_radii, axis=1)
    angles[:, 1::2] += np.pi/n_angles

    x = (radii*np.cos(angles)).flatten()
    y = (radii*np.sin(angles)).flatten()
    z = (np.cos(radii)*np.cos(angles*3.0)).flatten()

    # Create the Triangulation; no triangles so Delaunay triangulation created.
    triang = mtri.Triangulation(x, y)

    # Mask off unwanted triangles.
    xmid = x[triang.triangles].mean(axis=1)
    ymid = y[triang.triangles].mean(axis=1)
    mask = np.where(xmid*xmid + ymid*ymid < min_radius*min_radius, 1, 0)
    triang.set_mask(mask)

    v.isec(1)
    v.threed('on')
    v.trisurf(triang, z, cmap=plt.cm.CMRmap, cz=True, shade='linear',
              edgecolor=None)


def quiver_demo(**kwargs):
    '''
    quiver_demo(length = 1.0,
                normalize = True,
                facecolor = 'b',
                edgecolor = None,
                arrow_length_ratio = 0.3,
                shaftsize = 0.01,
                headsize = 0.01)

    '''
    from ifigure.interactive import quiver, threed, figure, lighting, view, isec, nsec

    x, y, z = np.meshgrid(np.arange(-0.8, 1, 0.2),
                          np.arange(-0.8, 1, 0.2),
                          np.arange(-0.8, 1, 0.8))

    u = np.sin(np.pi * x) * np.cos(np.pi * y) * np.cos(np.pi * z)
    v = -np.cos(np.pi * x) * np.sin(np.pi * y) * np.cos(np.pi * z)
    w = (np.sqrt(2.0 / 3.0) * np.cos(np.pi * x) * np.cos(np.pi * y) *
         np.sin(np.pi * z))

    figure()
    nsec(2)
    isec(0)
    threed('on')
    lighting(light=0.5, ambient=0.7)
    view('noclip')
    quiver(x, y, z, u, v, w, length=0.3, facecolor='b')
    isec(1)
    threed('on')
    lighting(light=0.5, ambient=0.7)
    view('noclip')
    quiver(x, y, z, u, v, w, cz=True, cdata=y, length=0.3)
